# -*- coding: utf-8 -*-
"""
Module for running arbitrary tests
"""
from __future__ import absolute_import, print_function, unicode_literals

# Import Python libs
import logging
import os
import random
import sys
import time
import traceback

# Import Salt libs
import salt.exceptions
import salt.loader
import salt.utils.args
import salt.utils.functools
import salt.utils.hashutils
import salt.utils.platform
import salt.version
from salt.ext import six
from salt.ext.six.moves import builtins
from salt.utils.decorators import depends

__proxyenabled__ = ["*"]

# Don't shadow built-in's.
__func_alias__ = {
    "true_": "true",
    "false_": "false",
    "try_": "try",
}

log = logging.getLogger(__name__)


@depends("non_existantmodulename")
def missing_func():
    return "foo"


def attr_call():
    """
    Call grains.items via the attribute

    CLI Example:

    .. code-block:: bash

        salt '*' test.attr_call
    """
    return __salt__.grains.items()


def module_report():
    """
    Return a dict containing all of the execution modules with a report on
    the overall availability via different references

    CLI Example:

    .. code-block:: bash

        salt '*' test.module_report
    """
    ret = {
        "functions": [],
        "function_attrs": [],
        "function_subs": [],
        "modules": [],
        "module_attrs": [],
        "missing_attrs": [],
        "missing_subs": [],
    }
    for ref in __salt__:
        if "." in ref:
            ret["functions"].append(ref)
        else:
            ret["modules"].append(ref)
            if hasattr(__salt__, ref):
                ret["module_attrs"].append(ref)
            for func in __salt__[ref]:
                full = "{0}.{1}".format(ref, func)
                if hasattr(getattr(__salt__, ref), func):
                    ret["function_attrs"].append(full)
                if func in __salt__[ref]:
                    ret["function_subs"].append(full)
    for func in ret["functions"]:
        if func not in ret["function_attrs"]:
            ret["missing_attrs"].append(func)
        if func not in ret["function_subs"]:
            ret["missing_subs"].append(func)
    return ret


def echo(text):
    """
    Return a string - used for testing the connection

    CLI Example:

    .. code-block:: bash

        salt '*' test.echo 'foo bar baz quo qux'
    """
    return text


def ping():
    """
    Used to make sure the minion is up and responding. Not an ICMP ping.

    Returns ``True``.

    CLI Example:

    .. code-block:: bash

        salt '*' test.ping
    """

    if not salt.utils.platform.is_proxy():
        log.debug("test.ping received for minion '%s'", __opts__.get("id"))
        return True
    else:
        ping_cmd = __opts__["proxy"]["proxytype"] + ".ping"
        if __opts__.get("add_proxymodule_to_opts", False):
            return __opts__["proxymodule"][ping_cmd]()
        else:
            return __proxy__[ping_cmd]()


def sleep(length):
    """
    Instruct the minion to initiate a process that will sleep for a given
    period of time.

    CLI Example:

    .. code-block:: bash

        salt '*' test.sleep 20
    """
    time.sleep(int(length))
    return True


def rand_sleep(max=60):
    """
    Sleep for a random number of seconds, used to test long-running commands
    and minions returning at differing intervals

    CLI Example:

    .. code-block:: bash

        salt '*' test.rand_sleep 60
    """
    time.sleep(random.randint(0, max))
    return True


def version():
    """
    Return the version of salt on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' test.version
    """
    return salt.version.__version__


def versions_information():
    """
    Report the versions of dependent and system software

    CLI Example:

    .. code-block:: bash

        salt '*' test.versions_information
    """
    return salt.version.versions_information()


def versions_report():
    """
    Returns versions of components used by salt

    CLI Example:

    .. code-block:: bash

        salt '*' test.versions_report
    """
    return "\n".join(salt.version.versions_report())


versions = salt.utils.functools.alias_function(versions_report, "versions")


def conf_test():
    """
    Return the value for test.foo in the minion configuration file, or return
    the default value

    CLI Example:

    .. code-block:: bash

        salt '*' test.conf_test
    """
    return __salt__["config.option"]("test.foo")


def get_opts():
    """
    Return the configuration options passed to this minion

    CLI Example:

    .. code-block:: bash

        salt '*' test.get_opts
    """
    return __opts__


def cross_test(func, args=None):
    """
    Execute a minion function via the __salt__ object in the test
    module, used to verify that the minion functions can be called
    via the __salt__ module.

    CLI Example:

    .. code-block:: bash

        salt '*' test.cross_test file.gid_to_group 0
    """
    if args is None:
        args = []
    return __salt__[func](*args)


def kwarg(**kwargs):
    """
    Print out the data passed into the function ``**kwargs``, this is used to
    both test the publication data and cli kwarg passing, but also to display
    the information available within the publication data.

    CLI Example:

    .. code-block:: bash

        salt '*' test.kwarg num=1 txt="two" env='{a: 1, b: "hello"}'
    """
    return kwargs


def arg(*args, **kwargs):
    """
    Print out the data passed into the function ``*args`` and ```kwargs``, this
    is used to both test the publication data and cli argument passing, but
    also to display the information available within the publication data.
    Returns {"args": args, "kwargs": kwargs}.

    CLI Example:

    .. code-block:: bash

        salt '*' test.arg 1 "two" 3.1 txt="hello" wow='{a: 1, b: "hello"}'
    """
    return {"args": args, "kwargs": kwargs}


def arg_type(*args, **kwargs):
    """
    Print out the types of the args and kwargs. This is used to test the types
    of the args and kwargs passed down to the minion

    CLI Example:

    .. code-block:: bash

           salt '*' test.arg_type 1 'int'
    """
    ret = {"args": [], "kwargs": {}}
    # all the args
    for argument in args:
        ret["args"].append(six.text_type(type(argument)))

    # all the kwargs
    for key, val in six.iteritems(kwargs):
        ret["kwargs"][key] = six.text_type(type(val))

    return ret


def arg_repr(*args, **kwargs):
    """
    Print out the data passed into the function ``*args`` and ```kwargs``, this
    is used to both test the publication data and cli argument passing, but
    also to display the information available within the publication data.
    Returns {"args": repr(args), "kwargs": repr(kwargs)}.

    CLI Example:

    .. code-block:: bash

        salt '*' test.arg_repr 1 "two" 3.1 txt="hello" wow='{a: 1, b: "hello"}'
    """
    return {"args": repr(args), "kwargs": repr(kwargs)}


def arg_clean(*args, **kwargs):
    """
    Like test.arg but cleans kwargs of the __pub* items
    CLI Example:

    .. code-block:: bash

        salt '*' test.arg_clean 1 "two" 3.1 txt="hello" wow='{a: 1, b: "hello"}'
    """
    return dict(args=args, kwargs=salt.utils.args.clean_kwargs(**kwargs))


def fib(num):
    """
    Return the num-th Fibonacci number, and the time it took to compute in
    seconds. Used for performance tests.

    This function is designed to have terrible performance.

    CLI Example:

    .. code-block:: bash

        salt '*' test.fib 3
    """
    num = int(num)
    if num < 0:
        raise ValueError("Negative number is not allowed!")
    start = time.time()
    if num < 2:
        return num, time.time() - start

    prev = 0
    curr = 1
    i = 1
    while i < num:
        prev, curr = curr, prev + curr
        i += 1

    return curr, time.time() - start


def collatz(start):
    """
    Execute the collatz conjecture from the passed starting number,
    returns the sequence and the time it took to compute. Used for
    performance tests.

    CLI Example:

    .. code-block:: bash

        salt '*' test.collatz 3
    """
    start = int(start)
    begin = time.time()
    steps = []
    while start != 1:
        steps.append(start)
        if start > 1:
            if start % 2 == 0:
                start = start / 2
            else:
                start = start * 3 + 1
    return steps, time.time() - begin


def outputter(data):
    """
    Test the outputter, pass in data to return

    CLI Example:

    .. code-block:: bash

        salt '*' test.outputter foobar
    """
    return data


def retcode(code=42):
    """
    Test that the returncode system is functioning correctly

    CLI Example:

    .. code-block:: bash

        salt '*' test.retcode 42
    """
    __context__["retcode"] = code
    return True


def provider(module):
    """
    Pass in a function name to discover what provider is being used

    CLI Example:

    .. code-block:: bash

        salt '*' test.provider service
    """
    func = ""
    for key in __salt__:
        if not key.startswith("{0}.".format(module)):
            continue
        func = key
        break
    if not func:
        return ""
    pfn = sys.modules[__salt__[func].__module__].__file__
    pfn = os.path.basename(pfn)
    return pfn[: pfn.rindex(".")]


def providers():
    """
    Return a dict of the provider names and the files that provided them

    CLI Example:

    .. code-block:: bash

        salt '*' test.providers
    """
    ret = {}
    for funcname in __salt__:
        modname = funcname.split(".")[0]
        if modname not in ret:
            ret[provider(modname)] = modname
    return ret


def not_loaded():
    """
    List the modules that were not loaded by the salt loader system

    CLI Example:

    .. code-block:: bash

        salt '*' test.not_loaded
    """
    prov = providers()
    ret = set()
    for mod_dir in salt.loader._module_dirs(__opts__, "modules", "module"):
        if not os.path.isabs(mod_dir):
            continue
        if not os.path.isdir(mod_dir):
            continue
        for fn_ in os.listdir(mod_dir):
            if fn_.startswith("_"):
                continue
            name = fn_.split(".")[0]
            if name not in prov:
                ret.add(name)
    return sorted(ret)


def opts_pkg():
    """
    Return an opts package with the grains and opts for this minion.
    This is primarily used to create the options used for master side
    state compiling routines

    CLI Example:

    .. code-block:: bash

        salt '*' test.opts_pkg
    """
    ret = {}
    ret.update(__opts__)
    ret["grains"] = __grains__
    return ret


def rand_str(size=9999999999, hash_type=None):
    """
    This function has been renamed to
    random_hash. This function will stay to
    ensure backwards compatibility, but please
    switch to using the preferred name random_hash.
    """
    return random_hash(size=size, hash_type=hash_type)


def random_hash(size=9999999999, hash_type=None):
    """
    .. versionadded:: 2015.5.2
    .. versionchanged:: 2018.3.0
        Function has been renamed from ``test.rand_str`` to
        ``test.random_hash``

    Generates a random number between 1 and ``size``, then returns a hash of
    that number. If no ``hash_type`` is passed, the hash_type specified by the
    minion's :conf_minion:`hash_type` config option is used.

    CLI Example:

    .. code-block:: bash

        salt '*' test.random_hash
        salt '*' test.random_hash hash_type=sha512
    """
    if not hash_type:
        hash_type = __opts__.get("hash_type", "md5")
    return salt.utils.hashutils.random_hash(size=size, hash_type=hash_type)


def exception(message="Test Exception"):
    """
    Raise an exception

    Optionally provide an error message or output the full stack.

    CLI Example:

    .. code-block:: bash

        salt '*' test.exception 'Oh noes!'
    """
    raise Exception(message)


def stack():
    """
    Return the current stack trace

    CLI Example:

    .. code-block:: bash

        salt '*' test.stack
    """
    return "".join(traceback.format_stack())


def tty(*args, **kwargs):  # pylint: disable=W0613
    """
    Deprecated! Moved to cmdmod.

    CLI Example:

    .. code-block:: bash

        salt '*' test.tty tty0 'This is a test'
        salt '*' test.tty pts3 'This is a test'
    """
    return "ERROR: This function has been moved to cmd.tty"


def try_(module, return_try_exception=False, **kwargs):
    """
    Try to run a module command. On an exception return None.
    If `return_try_exception` is set True return the exception.
    This can be helpful in templates where running a module might fail as expected.

    CLI Example:

    .. code-block:: bash

        <pre>
        {% for i in range(0,230) %}
            {{ salt['test.try'](module='ipmi.get_users', bmc_host='172.2.2.'+i)|yaml(False) }}
        {% endfor %}
        </pre>
    """
    try:
        return __salt__[module](**kwargs)
    except Exception as e:  # pylint: disable=broad-except
        if return_try_exception:
            return e
    return None


def assertion(assertion):
    """
    Assert the given argument

    CLI Example:

    .. code-block:: bash

        salt '*' test.assertion False
    """
    assert assertion


def true_():
    """
    Always return True

    CLI Example:

    .. code-block:: bash

        salt '*' test.true
    """
    return True


def false_():
    """
    Always return False

    CLI Example:

    .. code-block:: bash

        salt '*' test.false
    """
    return False


def raise_exception(name, *args, **kwargs):
    """
    Raise an exception. Built-in exceptions and those in ``salt.exceptions``
    can be raised by this test function. If no matching exception is found,
    then no exception will be raised and this function will return ``False``.

    This function is designed to test Salt's exception and return code
    handling.

    CLI Example:

    .. code-block:: bash

        salt '*' test.raise_exception TypeError "An integer is required"
        salt '*' test.raise_exception salt.exceptions.CommandExecutionError "Something went wrong"
    """

    def _is_exc(cls):
        for base in cls.__bases__:
            if base is BaseException:
                break
            else:
                return _is_exc(base)
        else:
            return False
        return True

    try:
        if name.startswith("salt.exceptions."):
            exc = getattr(salt.exceptions, name[16:])
        else:
            exc = getattr(builtins, name)
        if _is_exc(exc):
            raise exc(*args, **salt.utils.args.clean_kwargs(**kwargs))
        else:
            log.error("%s is not an exception", name)
            return False
    except AttributeError:
        log.error("No such exception: %s", name)
        return False



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.zk_concurrency as zk_concurrency

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class ZkConcurrencyTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate the zk_concurrency state
    """

    def setup_loader_modules(self):
        return {zk_concurrency: {}}

    def test_lock(self):
        """
        Test to block state execution until you are able to get the lock
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

        with patch.dict(zk_concurrency.__opts__, {"test": True}):
            ret.update({"comment": "Attempt to acquire lock", "result": None})
            self.assertDictEqual(zk_concurrency.lock("salt", "dude"), ret)

        with patch.dict(zk_concurrency.__opts__, {"test": False}):
            mock = MagicMock(return_value=True)
            with patch.dict(zk_concurrency.__salt__, {"zk_concurrency.lock": mock}):
                ret.update({"comment": "lock acquired", "result": True})
                self.assertDictEqual(zk_concurrency.lock("salt", "dude", "stack"), ret)

    def test_unlock(self):
        """
        Test to remove lease from semaphore
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

        with patch.dict(zk_concurrency.__opts__, {"test": True}):
            ret.update({"comment": "Released lock if it is here", "result": None})
            self.assertDictEqual(zk_concurrency.unlock("salt"), ret)

        with patch.dict(zk_concurrency.__opts__, {"test": False}):
            mock = MagicMock(return_value=True)
            with patch.dict(zk_concurrency.__salt__, {"zk_concurrency.unlock": mock}):
                ret.update({"comment": "", "result": True})
                self.assertDictEqual(
                    zk_concurrency.unlock("salt", identifier="stack"), ret
                )

    def test_min_party(self):
        """
        Test to ensure min party of nodes and the blocking behavior
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

        with patch.dict(zk_concurrency.__opts__, {"test": True}):
            ret.update({"comment": "Attempt to ensure min_party", "result": None})
            self.assertDictEqual(zk_concurrency.min_party("salt", "dude", 1), ret)

        with patch.dict(zk_concurrency.__opts__, {"test": False}):
            mock = MagicMock(return_value=["1", "2", "3"])
            with patch.dict(
                zk_concurrency.__salt__, {"zk_concurrency.party_members": mock}
            ):
                ret.update(
                    {"comment": "Currently 3 nodes, which is >= 2", "result": True}
                )
                self.assertDictEqual(zk_concurrency.min_party("salt", "dude", 2), ret)
                ret.update(
                    {
                        "comment": "Blocked until 2 nodes were available. "
                        + "Unblocked after 3 nodes became available",
                        "result": True,
                    }
                )
                self.assertDictEqual(
                    zk_concurrency.min_party("salt", "dude", 2, True), ret
                )
                ret.update(
                    {"comment": "Currently 3 nodes, which is < 4", "result": False}
                )
                self.assertDictEqual(zk_concurrency.min_party("salt", "dude", 4), ret)



# -*- coding: utf-8 -*-
"""
Device-Mapper module
"""
from __future__ import absolute_import, print_function, unicode_literals

import os.path


def multipath_list():
    """
    Device-Mapper Multipath list

    CLI Example:

    .. code-block:: bash

        salt '*' devmap.multipath_list
    """
    cmd = "multipath -l"
    return __salt__["cmd.run"](cmd).splitlines()


def multipath_flush(device):
    """
    Device-Mapper Multipath flush

    CLI Example:

    .. code-block:: bash

        salt '*' devmap.multipath_flush mpath1
    """
    if not os.path.exists(device):
        return "{0} does not exist".format(device)

    cmd = "multipath -f {0}".format(device)
    return __salt__["cmd.run"](cmd).splitlines()



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

import sys

# Import Salt Libs
import salt.states.macpackage as macpackage

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


@skipIf(sys.platform.startswith("win"), "Not a Windows test")
class MacPackageTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {macpackage: {}}

    def test_installed_pkg(self):
        """
            Test installing a PKG file
        """
        expected = {
            "changes": {"installed": ["some.other.id"]},
            "comment": "/path/to/file.pkg installed",
            "name": "/path/to/file.pkg",
            "result": True,
        }

        installed_mock = MagicMock(return_value=["com.apple.id"])
        get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
        install_mock = MagicMock(return_value={"retcode": 0})

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.installed_pkgs": installed_mock,
                "macpackage.get_pkg_id": get_pkg_id_mock,
                "macpackage.install": install_mock,
            },
        ):
            out = macpackage.installed("/path/to/file.pkg")
            installed_mock.assert_called_once_with()
            get_pkg_id_mock.assert_called_once_with("/path/to/file.pkg")
            install_mock.assert_called_once_with(
                "/path/to/file.pkg", "LocalSystem", False, False
            )
            self.assertEqual(out, expected)

    def test_installed_pkg_exists(self):
        """
            Test installing a PKG file where it's already installed
        """
        expected = {
            "changes": {},
            "comment": "",
            "name": "/path/to/file.pkg",
            "result": True,
        }

        installed_mock = MagicMock(return_value=["com.apple.id", "some.other.id"])
        get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
        install_mock = MagicMock(return_value={"retcode": 0})

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.installed_pkgs": installed_mock,
                "macpackage.get_pkg_id": get_pkg_id_mock,
                "macpackage.install": install_mock,
            },
        ):
            out = macpackage.installed("/path/to/file.pkg")
            installed_mock.assert_called_once_with()
            get_pkg_id_mock.assert_called_once_with("/path/to/file.pkg")
            assert not install_mock.called
            self.assertEqual(out, expected)

    def test_installed_pkg_version_succeeds(self):
        """
            Test installing a PKG file where the version number matches the current installed version
        """
        expected = {
            "changes": {},
            "comment": "Version already matches .*5\\.1\\.[0-9]",
            "name": "/path/to/file.pkg",
            "result": True,
        }

        installed_mock = MagicMock(return_value=["com.apple.id", "some.other.id"])
        get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
        install_mock = MagicMock(return_value={"retcode": 0})
        cmd_mock = MagicMock(return_value="Version of this: 5.1.9")

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.installed_pkgs": installed_mock,
                "macpackage.get_pkg_id": get_pkg_id_mock,
                "macpackage.install": install_mock,
                "cmd.run": cmd_mock,
            },
        ):
            out = macpackage.installed(
                "/path/to/file.pkg",
                version_check=r"/usr/bin/runme --version=.*5\.1\.[0-9]",
            )
            cmd_mock.assert_called_once_with(
                "/usr/bin/runme --version", output_loglevel="quiet", ignore_retcode=True
            )
            assert not installed_mock.called
            assert not get_pkg_id_mock.called
            assert not install_mock.called
            self.assertEqual(out, expected)

    def test_installed_pkg_version_fails(self):
        """
            Test installing a PKG file where the version number if different from the expected one
        """
        expected = {
            "changes": {"installed": ["some.other.id"]},
            "comment": "Version Version of this: 1.8.9 doesn't match .*5\\.1\\.[0-9]. /path/to/file.pkg installed",
            "name": "/path/to/file.pkg",
            "result": True,
        }

        installed_mock = MagicMock(return_value=["com.apple.id"])
        get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
        install_mock = MagicMock(return_value={"retcode": 0})
        cmd_mock = MagicMock(return_value="Version of this: 1.8.9")

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.installed_pkgs": installed_mock,
                "macpackage.get_pkg_id": get_pkg_id_mock,
                "macpackage.install": install_mock,
                "cmd.run": cmd_mock,
            },
        ):
            out = macpackage.installed(
                "/path/to/file.pkg",
                version_check=r"/usr/bin/runme --version=.*5\.1\.[0-9]",
            )
            cmd_mock.assert_called_once_with(
                "/usr/bin/runme --version", output_loglevel="quiet", ignore_retcode=True
            )
            installed_mock.assert_called_once_with()
            get_pkg_id_mock.assert_called_once_with("/path/to/file.pkg")
            install_mock.assert_called_once_with(
                "/path/to/file.pkg", "LocalSystem", False, False
            )
            self.assertEqual(out, expected)

    def test_installed_dmg(self):
        """
            Test installing a DMG file
        """
        expected = {
            "changes": {"installed": ["some.other.id"]},
            "comment": "/path/to/file.dmg installed",
            "name": "/path/to/file.dmg",
            "result": True,
        }

        mount_mock = MagicMock(return_value=["success", "/tmp/dmg-X"])
        unmount_mock = MagicMock()
        installed_mock = MagicMock(return_value=["com.apple.id"])
        get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
        install_mock = MagicMock(return_value={"retcode": 0})

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.mount": mount_mock,
                "macpackage.unmount": unmount_mock,
                "macpackage.installed_pkgs": installed_mock,
                "macpackage.get_pkg_id": get_pkg_id_mock,
                "macpackage.install": install_mock,
            },
        ):
            out = macpackage.installed("/path/to/file.dmg", dmg=True)
            mount_mock.assert_called_once_with("/path/to/file.dmg")
            unmount_mock.assert_called_once_with("/tmp/dmg-X")
            installed_mock.assert_called_once_with()
            get_pkg_id_mock.assert_called_once_with("/tmp/dmg-X/*.pkg")
            install_mock.assert_called_once_with(
                "/tmp/dmg-X/*.pkg", "LocalSystem", False, False
            )
            self.assertEqual(out, expected)

    def test_installed_dmg_exists(self):
        """
            Test installing a DMG file when the package already exists
        """
        expected = {
            "changes": {},
            "comment": "",
            "name": "/path/to/file.dmg",
            "result": True,
        }

        mount_mock = MagicMock(return_value=["success", "/tmp/dmg-X"])
        unmount_mock = MagicMock()
        installed_mock = MagicMock(return_value=["com.apple.id", "some.other.id"])
        get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
        install_mock = MagicMock(return_value={"retcode": 0})

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.mount": mount_mock,
                "macpackage.unmount": unmount_mock,
                "macpackage.installed_pkgs": installed_mock,
                "macpackage.get_pkg_id": get_pkg_id_mock,
                "macpackage.install": install_mock,
            },
        ):
            out = macpackage.installed("/path/to/file.dmg", dmg=True)
            mount_mock.assert_called_once_with("/path/to/file.dmg")
            unmount_mock.assert_called_once_with("/tmp/dmg-X")
            installed_mock.assert_called_once_with()
            get_pkg_id_mock.assert_called_once_with("/tmp/dmg-X/*.pkg")
            assert not install_mock.called
            self.assertEqual(out, expected)

    def test_installed_app(self):
        """
            Test installing an APP file
        """
        with patch("os.path.exists") as exists_mock:
            expected = {
                "changes": {"installed": ["file.app"]},
                "comment": "file.app installed",
                "name": "/path/to/file.app",
                "result": True,
            }

            install_mock = MagicMock()
            exists_mock.return_value = False

            with patch.dict(
                macpackage.__salt__, {"macpackage.install_app": install_mock}
            ):
                out = macpackage.installed("/path/to/file.app", app=True)

                install_mock.assert_called_once_with(
                    "/path/to/file.app", "/Applications/"
                )
                self.assertEqual(out, expected)

    def test_installed_app_exists(self):
        """
            Test installing an APP file that already exists
        """
        with patch("os.path.exists") as exists_mock:
            expected = {
                "changes": {},
                "comment": "",
                "name": "/path/to/file.app",
                "result": True,
            }

            install_mock = MagicMock()
            exists_mock.return_value = True

            with patch.dict(
                macpackage.__salt__, {"macpackage.install_app": install_mock}
            ):
                out = macpackage.installed("/path/to/file.app", app=True)

                assert not install_mock.called
                self.assertEqual(out, expected)

    def test_installed_app_dmg(self):
        """
            Test installing an APP file contained in a DMG file
        """
        with patch("os.path.exists") as exists_mock:
            expected = {
                "changes": {"installed": ["file.app"]},
                "comment": "file.app installed",
                "name": "/path/to/file.dmg",
                "result": True,
            }

            install_mock = MagicMock()
            mount_mock = MagicMock(return_value=["success", "/tmp/dmg-X"])
            unmount_mock = MagicMock()
            cmd_mock = MagicMock(return_value="file.app")
            exists_mock.return_value = False

            with patch.dict(
                macpackage.__salt__,
                {
                    "macpackage.install_app": install_mock,
                    "macpackage.mount": mount_mock,
                    "macpackage.unmount": unmount_mock,
                    "cmd.run": cmd_mock,
                },
            ):
                out = macpackage.installed("/path/to/file.dmg", app=True, dmg=True)

                mount_mock.assert_called_once_with("/path/to/file.dmg")
                unmount_mock.assert_called_once_with("/tmp/dmg-X")
                cmd_mock.assert_called_once_with(
                    "ls -d *.app", python_shell=True, cwd="/tmp/dmg-X"
                )
                install_mock.assert_called_once_with(
                    "/tmp/dmg-X/file.app", "/Applications/"
                )
                self.assertEqual(out, expected)

    def test_installed_app_dmg_exists(self):
        """
            Test installing an APP file contained in a DMG file where the file exists
        """
        with patch("os.path.exists") as exists_mock:
            expected = {
                "changes": {},
                "comment": "",
                "name": "/path/to/file.dmg",
                "result": True,
            }

            install_mock = MagicMock()
            mount_mock = MagicMock(return_value=["success", "/tmp/dmg-X"])
            unmount_mock = MagicMock()
            cmd_mock = MagicMock(return_value="file.app")
            exists_mock.return_value = True

            with patch.dict(
                macpackage.__salt__,
                {
                    "macpackage.install_app": install_mock,
                    "macpackage.mount": mount_mock,
                    "macpackage.unmount": unmount_mock,
                    "cmd.run": cmd_mock,
                },
            ):
                out = macpackage.installed("/path/to/file.dmg", app=True, dmg=True)

                mount_mock.assert_called_once_with("/path/to/file.dmg")
                unmount_mock.assert_called_once_with("/tmp/dmg-X")
                cmd_mock.assert_called_once_with(
                    "ls -d *.app", python_shell=True, cwd="/tmp/dmg-X"
                )
                assert not install_mock.called
                self.assertEqual(out, expected)

    def test_installed_pkg_only_if_pass(self):
        """
            Test installing a PKG file where the only if call passes
        """
        expected = {
            "changes": {"installed": ["some.other.id"]},
            "comment": "/path/to/file.pkg installed",
            "name": "/path/to/file.pkg",
            "result": True,
        }

        installed_mock = MagicMock(return_value=["com.apple.id"])
        get_pkg_id_mock = MagicMock(return_value=["some.other.id"])
        install_mock = MagicMock(return_value={"retcode": 0})
        cmd_mock = MagicMock(return_value=0)

        with patch.dict(
            macpackage.__salt__,
            {
                "macpackage.installed_pkgs": installed_mock,
                "macpackage.get_pkg_id": get_pkg_id_mock,
                "macpackage.install": install_mock,
                "cmd.retcode": cmd_mock,
            },
        ):
            out = macpackage.installed("/path/to/file.pkg")
            installed_mock.assert_called_once_with()
            get_pkg_id_mock.assert_called_once_with("/path/to/file.pkg")
            install_mock.assert_called_once_with(
                "/path/to/file.pkg", "LocalSystem", False, False
            )
            self.assertEqual(out, expected)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.modules.sdb as sdb

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


class SdbTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.sdb
    """

    def setup_loader_modules(self):
        return {sdb: {}}

    # 'get' function tests: 1

    def test_get(self):
        """
        Test if it gets a value from a db, using a uri in the form of
        sdb://<profile>/<key>
        """
        self.assertEqual(sdb.get("sdb://salt/foo"), "sdb://salt/foo")

    # 'set_' function tests: 1

    def test_set(self):
        """
        Test if it sets a value from a db, using a uri in the form of
        sdb://<profile>/<key>
        """
        self.assertFalse(sdb.set_("sdb://mymemcached/foo", "bar"))



# -*- coding: utf-8 -*-

"""
tests for pkg state
"""
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import time

import pytest
import salt.utils.files
import salt.utils.path
import salt.utils.pkg.rpm
import salt.utils.platform
from salt.ext import six
from salt.ext.six.moves import range
from tests.support.case import ModuleCase
from tests.support.helpers import (
    destructiveTest,
    not_runs_on,
    requires_salt_modules,
    requires_salt_states,
    requires_system_grains,
    runs_on,
    slowTest,
)
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.unit import skipIf

log = logging.getLogger(__name__)


@destructiveTest
@pytest.mark.windows_whitelisted
class PkgTest(ModuleCase, SaltReturnAssertsMixin):
    _PKG_EPOCH_TARGETS = []
    _PKG_32_TARGETS = []
    _PKG_CAP_TARGETS = []
    _PKG_DOT_TARGETS = []
    _WILDCARDS_SUPPORTED = False
    _VERSION_SPEC_SUPPORTED = True

    @classmethod
    @requires_system_grains
    def setUpClass(cls, grains=None):  # pylint:disable=W0221
        cls.ctx = {}
        cls._PKG_TARGETS = ["figlet", "sl"]
        if grains["os"] == "Windows":
            cls._PKG_TARGETS = ["7zip", "putty"]
        elif grains["os"] == "FreeBSD":
            cls._VERSION_SPEC_SUPPORTED = False
        elif grains["os_family"] in ("Arch", "Debian"):
            cls._WILDCARDS_SUPPORTED = True
        elif grains["os"] == "Amazon":
            cls._PKG_TARGETS = ["lynx", "gnuplot"]
        elif grains["os_family"] == "RedHat":
            cls._PKG_TARGETS = ["units", "zsh-html"]
            cls._WILDCARDS_SUPPORTED = True
            if grains["os"] == "CentOS":
                if grains["osmajorrelease"] == 5:
                    cls._PKG_32_TARGETS = ["xz-devel.i386"]
                else:
                    cls._PKG_32_TARGETS.append("xz-devel.i686")
            if grains["osmajorrelease"] == 5:
                cls._PKG_DOT_TARGETS = ["python-migrate0.5"]
            elif grains["osmajorrelease"] == 6:
                cls._PKG_DOT_TARGETS = ["tomcat6-el-2.1-api"]
            elif grains["osmajorrelease"] == 7:
                cls._PKG_DOT_TARGETS = ["tomcat-el-2.2-api"]
                cls._PKG_EPOCH_TARGETS = ["comps-extras"]
            elif grains["osmajorrelease"] == 8:
                cls._PKG_DOT_TARGETS = ["vid.stab"]
                cls._PKG_EPOCH_TARGETS = ["traceroute"]
        elif grains["os_family"] == "Suse":
            cls._PKG_TARGETS = ["lynx", "htop"]
            if grains["os"] == "SUSE":
                cls._PKG_CAP_TARGETS = [("perl(ZNC)", "znc-perl")]

    @classmethod
    def tearDownClass(cls):
        del cls.ctx

    def latest_version(self, *names):
        """
        Helper function which ensures that we don't make any unnecessary calls to
        pkg.latest_version to figure out what version we need to install. This
        won't stop pkg.latest_version from being run in a pkg.latest state, but it
        will reduce the amount of times we check the latest version here in the
        test suite.
        """
        key = "latest_version"
        if key not in self.ctx:
            self.ctx[key] = dict()
        targets = [x for x in names if x not in self.ctx[key]]
        if targets:
            result = self.run_function("pkg.latest_version", targets, refresh=False)
            try:
                self.ctx[key].update(result)
            except ValueError:
                # Only a single target, pkg.latest_version returned a string
                self.ctx[key][targets[0]] = result

        ret = dict([(x, self.ctx[key].get(x, "")) for x in names])
        if len(names) == 1:
            return ret[names[0]]
        return ret

    @requires_system_grains
    def setUp(self, grains=None):  # pylint:disable=W0221
        super(PkgTest, self).setUp()
        if "refresh" not in self.ctx:
            self.run_function("pkg.refresh_db")
            self.ctx["refresh"] = True

        # If this is Arch Linux, check if pacman is in use by another process
        if grains["os_family"] == "Arch":
            for _ in range(12):
                if not os.path.isfile("/var/lib/pacman/db.lck"):
                    break
                else:
                    time.sleep(5)
            else:
                raise Exception("Package database locked after 60 seconds, bailing out")

    @requires_salt_modules("pkg.version")
    @requires_salt_states("pkg.installed", "pkg.removed")
    @slowTest
    def test_pkg_001_installed(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        target = self._PKG_TARGETS[0]
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertFalse(version)

        ret = self.run_state("pkg.installed", name=target, refresh=False)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_002_installed_with_version(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._VERSION_SPEC_SUPPORTED:
            self.skipTest("Version specification not supported")
        target = self._PKG_TARGETS[0]
        version = self.latest_version(target)

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertTrue(version)

        ret = self.run_state(
            "pkg.installed", name=target, version=version, refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_states("pkg.installed", "pkg.removed")
    @slowTest
    def test_pkg_003_installed_multipkg(self):
        """
        This is a destructive test as it installs and then removes two packages
        """
        version = self.run_function("pkg.version", self._PKG_TARGETS)

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so these
        # packages need to not be installed before we run the states below
        self.assertFalse(any(version.values()))
        self.assertSaltTrueReturn(
            self.run_state("pkg.removed", name=None, pkgs=self._PKG_TARGETS)
        )

        try:
            ret = self.run_state(
                "pkg.installed", name=None, pkgs=self._PKG_TARGETS, refresh=False
            )
            self.assertSaltTrueReturn(ret)
        finally:
            ret = self.run_state("pkg.removed", name=None, pkgs=self._PKG_TARGETS)
            self.assertSaltTrueReturn(ret)

    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_004_installed_multipkg_with_version(self):
        """
        This is a destructive test as it installs and then removes two packages
        """
        if not self._VERSION_SPEC_SUPPORTED:
            self.skipTest("Version specification not supported")
        version = self.latest_version(self._PKG_TARGETS[0])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so these
        # packages need to not be installed before we run the states below
        self.assertTrue(bool(version))

        pkgs = [{self._PKG_TARGETS[0]: version}, self._PKG_TARGETS[1]]

        try:
            ret = self.run_state("pkg.installed", name=None, pkgs=pkgs, refresh=False)
            self.assertSaltTrueReturn(ret)
        finally:
            ret = self.run_state("pkg.removed", name=None, pkgs=self._PKG_TARGETS)
            self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.version")
    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_005_installed_32bit(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_32_TARGETS:
            self.skipTest("No 32 bit packages have been specified for testing")

        target = self._PKG_32_TARGETS[0]

        # _PKG_TARGETS_32 is only populated for platforms for which Salt has to
        # munge package names for 32-bit-on-x86_64 (Currently only Ubuntu and
        # RHEL-based). Don't actually perform this test on other platforms.
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of packages, so
        # the target needs to not be installed before we run the states
        # below
        self.assertFalse(version)

        ret = self.run_state("pkg.installed", name=target, refresh=False)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_006_installed_32bit_with_version(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_32_TARGETS:
            self.skipTest("No 32 bit packages have been specified for testing")

        target = self._PKG_32_TARGETS[0]

        # _PKG_TARGETS_32 is only populated for platforms for which Salt has to
        # munge package names for 32-bit-on-x86_64 (Currently only Ubuntu and
        # RHEL-based). Don't actually perform this test on other platforms.
        version = self.latest_version(target)

        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of the package, so
        # the target needs to not be installed before we run the states
        # below
        self.assertTrue(version)

        ret = self.run_state(
            "pkg.installed", name=target, version=version, refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_007_with_dot_in_pkgname(self=None):
        """
        This tests for the regression found in the following issue:
        https://github.com/saltstack/salt/issues/8614

        This is a destructive test as it installs a package
        """
        if not self._PKG_DOT_TARGETS:
            self.skipTest('No packages with "." in their name have been specified',)

        target = self._PKG_DOT_TARGETS[0]

        version = self.latest_version(target)
        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of the package, so
        # the target needs to not be installed before we run the
        # pkg.installed state below
        self.assertTrue(bool(version))
        ret = self.run_state("pkg.installed", name=target, refresh=False)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_008_epoch_in_version(self):
        """
        This tests for the regression found in the following issue:
        https://github.com/saltstack/salt/issues/8614

        This is a destructive test as it installs a package
        """
        if not self._PKG_EPOCH_TARGETS:
            self.skipTest('No targets have been configured with "epoch" in the version')

        target = self._PKG_EPOCH_TARGETS[0]

        version = self.latest_version(target)
        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of the package, so
        # the target needs to not be installed before we run the
        # pkg.installed state below
        self.assertTrue(version)
        ret = self.run_state(
            "pkg.installed", name=target, version=version, refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.version", "pkg.info_installed")
    @requires_salt_states("pkg.installed", "pkg.removed")
    @runs_on(kernel="linux")
    @not_runs_on(os="Amazon")
    @slowTest
    def test_pkg_009_latest_with_epoch(self):
        """
        This tests for the following issue:
        https://github.com/saltstack/salt/issues/31014

        This is a destructive test as it installs a package
        """
        package = "bash-completion"
        pkgquery = "version"

        ret = self.run_state("pkg.installed", name=package, refresh=False)
        self.assertSaltTrueReturn(ret)

        ret = self.run_function("pkg.info_installed", [package])
        self.assertTrue(pkgquery in six.text_type(ret))

    @requires_salt_states("pkg.latest", "pkg.removed")
    @slowTest
    def test_pkg_010_latest(self):
        """
        This tests pkg.latest with a package that has no epoch (or a zero
        epoch).
        """
        target = self._PKG_TARGETS[0]
        version = self.latest_version(target)

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertTrue(version)

        ret = self.run_state("pkg.latest", name=target, refresh=False)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.list_pkgs", "pkg.list_upgrades", "pkg.version")
    @requires_salt_states("pkg.latest")
    @runs_on(kernel="linux", os_family="Debian")
    @slowTest
    def test_pkg_011_latest_only_upgrade(self):
        """
        WARNING: This test will pick a package with an available upgrade (if
        there is one) and upgrade it to the latest version.
        """
        target = self._PKG_TARGETS[0]

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test that the state fails when you try to run the state
        # with only_upgrade=True on a package which is not already installed,
        # so the targeted package needs to not be installed before we run the
        # state below.
        version = self.latest_version(target)
        self.assertTrue(version)

        ret = self.run_state(
            "pkg.latest", name=target, refresh=False, only_upgrade=True
        )
        self.assertSaltFalseReturn(ret)

        # Now look for updates and try to run the state on a package which is already up-to-date.
        installed_pkgs = self.run_function("pkg.list_pkgs")
        updates = self.run_function("pkg.list_upgrades", refresh=False)

        for pkgname in updates:
            if pkgname in installed_pkgs:
                target = pkgname
                break
        else:
            target = ""
            log.warning(
                "No available upgrades to installed packages, skipping "
                "only_upgrade=True test with already-installed package. For "
                "best results run this test on a machine with upgrades "
                "available."
            )

        if target:
            ret = self.run_state(
                "pkg.latest", name=target, refresh=False, only_upgrade=True
            )
            self.assertSaltTrueReturn(ret)
            new_version = self.run_function("pkg.version", [target])
            self.assertEqual(new_version, updates[target])
            ret = self.run_state(
                "pkg.latest", name=target, refresh=False, only_upgrade=True
            )
            self.assertEqual(
                ret["pkg_|-{0}_|-{0}_|-latest".format(target)]["comment"],
                "Package {0} is already up-to-date".format(target),
            )

    @requires_salt_modules("pkg.version")
    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_012_installed_with_wildcard_version(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._WILDCARDS_SUPPORTED:
            self.skipTest("Wildcards in pkg.install are not supported")

        target = self._PKG_TARGETS[0]
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertFalse(version)

        ret = self.run_state("pkg.installed", name=target, version="*", refresh=False,)
        self.assertSaltTrueReturn(ret)

        # Repeat state, should pass
        ret = self.run_state("pkg.installed", name=target, version="*", refresh=False,)

        expected_comment = (
            "All specified packages are already installed and are at the "
            "desired version"
        )
        self.assertSaltTrueReturn(ret)
        self.assertEqual(ret[next(iter(ret))]["comment"], expected_comment)

        # Repeat one more time with unavailable version, test should fail
        ret = self.run_state(
            "pkg.installed", name=target, version="93413*", refresh=False,
        )
        self.assertSaltFalseReturn(ret)

        # Clean up
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.version", "pkg.latest_version")
    @requires_salt_states("pkg.installed", "pkg.removed")
    @runs_on(kernel="linux", os_family=["Debian", "RedHat"])
    @slowTest
    def test_pkg_013_installed_with_comparison_operator(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        target = self._PKG_TARGETS[0]
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertFalse(version)

        latest_version = self.run_function(
            "pkg.latest_version", [target], refresh=False
        )

        try:
            ret = self.run_state(
                "pkg.installed", name=target, version="<9999999", refresh=False,
            )
            self.assertSaltTrueReturn(ret)

            # The version that was installed should be the latest available
            version = self.run_function("pkg.version", [target])
            self.assertTrue(version, latest_version)
        finally:
            # Clean up
            ret = self.run_state("pkg.removed", name=target)
            self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.version")
    @requires_salt_states("pkg.installed", "pkg.removed")
    @runs_on(kernel="linux", os_familiy="RedHat")
    def test_pkg_014_installed_missing_release(self):
        """
        Tests that a version number missing the release portion still resolves
        as correctly installed. For example, version 2.0.2 instead of 2.0.2-1.el7
        """
        target = self._PKG_TARGETS[0]
        version = self.run_function("pkg.version", [target])

        # If this assert fails, we need to find new targets, this test needs to
        # be able to test successful installation of packages, so this package
        # needs to not be installed before we run the states below
        self.assertFalse(version)

        ret = self.run_state(
            "pkg.installed",
            name=target,
            version=salt.utils.pkg.rpm.version_to_evr(version)[1],
            refresh=False,
        )
        self.assertSaltTrueReturn(ret)

        # Clean up
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.hold", "pkg.unhold", "pkg.version", "pkg.list_pkgs")
    @requires_salt_states("pkg.installed", "pkg.removed")
    @requires_system_grains
    @slowTest
    def test_pkg_015_installed_held(self, grains=None):
        """
        Tests that a package can be held even when the package is already installed.
        """
        versionlock_pkg = None
        if grains["os_family"] == "RedHat":
            pkgs = {
                p for p in self.run_function("pkg.list_pkgs") if "-versionlock" in p
            }
            if not pkgs:
                self.skipTest("No versionlock package found in repositories")
            for versionlock_pkg in pkgs:
                ret = self.run_state(
                    "pkg.installed", name=versionlock_pkg, refresh=False
                )
                # Exit loop if a versionlock package installed correctly
                try:
                    self.assertSaltTrueReturn(ret)
                    log.debug(
                        "Installed versionlock package: {}".format(versionlock_pkg)
                    )
                    break
                except AssertionError as e:
                    log.debug("Versionlock package not found:\n{}".format(e))
            else:
                self.fail("Could not install versionlock package from {}".format(pkgs))

        target = self._PKG_TARGETS[0]

        # First we ensure that the package is installed
        ret = self.run_state("pkg.installed", name=target, refresh=False,)
        self.assertSaltTrueReturn(ret)

        # Then we check that the package is now held
        ret = self.run_state("pkg.installed", name=target, hold=True, refresh=False,)

        if versionlock_pkg and "-versionlock is not installed" in str(ret):
            self.skipTest("{}  `{}` is installed".format(ret, versionlock_pkg))

        # changes from pkg.hold for Red Hat family are different
        target_changes = {}
        if grains["os_family"] == "RedHat" or grains["os"] == "FreeBSD":
            target_changes = {"new": "hold", "old": ""}
        elif grains["os_family"] == "Debian":
            target_changes = {"new": "hold", "old": "install"}

        try:
            tag = "pkg_|-{0}_|-{0}_|-installed".format(target)
            self.assertSaltTrueReturn(ret)
            self.assertIn(tag, ret)
            self.assertIn("changes", ret[tag])
            self.assertIn(target, ret[tag]["changes"])
            if not target_changes:
                self.skipTest(
                    "Test needs to be configured for {}: {}".format(
                        grains["os"], ret[tag]["changes"][target]
                    )
                )
            self.assertEqual(ret[tag]["changes"][target], target_changes)
        finally:
            # Clean up, unhold package and remove
            self.run_function("pkg.unhold", name=target)
            ret = self.run_state("pkg.removed", name=target)
            self.assertSaltTrueReturn(ret)
            if versionlock_pkg:
                ret = self.run_state("pkg.removed", name=versionlock_pkg)
                self.assertSaltTrueReturn(ret)

    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_016_conditionally_ignore_epoch(self):
        """
        See
        https://github.com/saltstack/salt/issues/56654#issuecomment-615034952

        This is a destructive test as it installs a package
        """
        if not self._PKG_EPOCH_TARGETS:
            self.skipTest('No targets have been configured with "epoch" in the version')

        target = self._PKG_EPOCH_TARGETS[0]

        # Strip the epoch from the latest available version
        version = self.latest_version(target).split(":", 1)[-1]
        # If this assert fails, we need to find a new target. This test
        # needs to be able to test successful installation of the package, so
        # the target needs to not be installed before we run the
        # pkg.installed state below
        self.assertTrue(version)

        # CASE 1: package name passed in "name" param
        ret = self.run_state(
            "pkg.installed", name=target, version=version, refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

        # CASE 2: same as case 1 but with "pkgs"
        ret = self.run_state(
            "pkg.installed", name="foo", pkgs=[{target: version}], refresh=False
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("pkg.removed", name=target)
        self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.version")
    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_cap_001_installed(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        try:
            ret = self.run_state(
                "pkg.installed",
                name=target,
                refresh=False,
                resolve_capabilities=True,
                test=True,
            )
            self.assertInSaltComment(
                "The following packages would be installed/updated: {0}".format(
                    realpkg
                ),
                ret,
            )
            ret = self.run_state(
                "pkg.installed", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)
        finally:
            ret = self.run_state("pkg.removed", name=realpkg)
            self.assertSaltTrueReturn(ret)

    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_cap_002_already_installed(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        try:
            # install the package
            ret = self.run_state("pkg.installed", name=realpkg, refresh=False)
            self.assertSaltTrueReturn(ret)

            # Try to install again. Nothing should be installed this time.
            ret = self.run_state(
                "pkg.installed",
                name=target,
                refresh=False,
                resolve_capabilities=True,
                test=True,
            )
            self.assertInSaltComment(
                "All specified packages are already installed", ret
            )

            ret = self.run_state(
                "pkg.installed", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)

            self.assertInSaltComment("packages are already installed", ret)
        finally:
            ret = self.run_state("pkg.removed", name=realpkg)
            self.assertSaltTrueReturn(ret)

    @skipIf(not _PKG_CAP_TARGETS, "Capability not available")
    @requires_salt_states("pkg.installed", "pkg.removed")
    def test_pkg_cap_003_installed_multipkg_with_version(self):
        """
        This is a destructive test as it installs and then removes two packages
        """
        if not self._VERSION_SPEC_SUPPORTED:
            self.skipTest("Version specification not supported")
        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.latest_version(target)
        realver = self.latest_version(realpkg)

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        cleanup_pkgs = self._PKG_TARGETS
        try:
            pkgs = [
                {self._PKG_TARGETS[0]: version},
                self._PKG_TARGETS[1],
                {target: realver},
            ]
            ret = self.run_state(
                "pkg.installed",
                name="test_pkg_cap_003_installed_multipkg_with_version-install",
                pkgs=pkgs,
                refresh=False,
            )
            self.assertSaltFalseReturn(ret)

            ret = self.run_state(
                "pkg.installed",
                name="test_pkg_cap_003_installed_multipkg_with_version-install-capability",
                pkgs=pkgs,
                refresh=False,
                resolve_capabilities=True,
                test=True,
            )
            self.assertInSaltComment("packages would be installed/updated", ret)
            self.assertInSaltComment("{0}={1}".format(realpkg, realver), ret)

            ret = self.run_state(
                "pkg.installed",
                name="test_pkg_cap_003_installed_multipkg_with_version-install-capability",
                pkgs=pkgs,
                refresh=False,
                resolve_capabilities=True,
            )
            self.assertSaltTrueReturn(ret)
            cleanup_pkgs.append(realpkg)
        finally:
            ret = self.run_state(
                "pkg.removed",
                name="test_pkg_cap_003_installed_multipkg_with_version-remove",
                pkgs=cleanup_pkgs,
            )
            self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.version")
    @requires_salt_states("pkg.latest", "pkg.removed")
    def test_pkg_cap_004_latest(self):
        """
        This tests pkg.latest with a package that has no epoch (or a zero
        epoch).
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        try:
            ret = self.run_state(
                "pkg.latest",
                name=target,
                refresh=False,
                resolve_capabilities=True,
                test=True,
            )
            self.assertInSaltComment(
                "The following packages would be installed/upgraded: {0}".format(
                    realpkg
                ),
                ret,
            )
            ret = self.run_state(
                "pkg.latest", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)

            ret = self.run_state(
                "pkg.latest", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)
            self.assertInSaltComment("is already up-to-date", ret)
        finally:
            ret = self.run_state("pkg.removed", name=realpkg)
            self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.version")
    @requires_salt_states("pkg.installed", "pkg.removed", "pkg.downloaded")
    def test_pkg_cap_005_downloaded(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        ret = self.run_state("pkg.downloaded", name=target, refresh=False)
        self.assertSaltFalseReturn(ret)

        ret = self.run_state(
            "pkg.downloaded",
            name=target,
            refresh=False,
            resolve_capabilities=True,
            test=True,
        )
        self.assertInSaltComment(
            "The following packages would be downloaded: {0}".format(realpkg), ret
        )

        ret = self.run_state(
            "pkg.downloaded", name=target, refresh=False, resolve_capabilities=True
        )
        self.assertSaltTrueReturn(ret)

    @requires_salt_modules("pkg.version")
    @requires_salt_states("pkg.installed", "pkg.removed", "pkg.uptodate")
    def test_pkg_cap_006_uptodate(self):
        """
        This is a destructive test as it installs and then removes a package
        """
        if not self._PKG_CAP_TARGETS:
            self.skipTest("Capability not provided")

        target, realpkg = self._PKG_CAP_TARGETS[0]
        version = self.run_function("pkg.version", [target])
        realver = self.run_function("pkg.version", [realpkg])

        # If this condition is False, we need to find new targets.
        # This needs to be able to test successful installation of packages.
        # These packages need to not be installed before we run the states below
        if not (version and realver):
            self.skipTest("TODO: New pkg cap targets required")

        try:
            ret = self.run_state(
                "pkg.installed", name=target, refresh=False, resolve_capabilities=True
            )
            self.assertSaltTrueReturn(ret)
            ret = self.run_state(
                "pkg.uptodate",
                name="test_pkg_cap_006_uptodate",
                pkgs=[target],
                refresh=False,
                resolve_capabilities=True,
            )
            self.assertSaltTrueReturn(ret)
            self.assertInSaltComment("System is already up-to-date", ret)
        finally:
            ret = self.run_state("pkg.removed", name=realpkg)
            self.assertSaltTrueReturn(ret)



# -*- coding: utf-8 -*-
#
# Copyright 2015 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Import Python Libs
from __future__ import absolute_import

import logging
import os
import time

# Import Salt Libs
import salt.utils.files
import salt.utils.fsutils
import salt.utils.network
from salt.modules.inspectlib import EnvLoader
from salt.modules.inspectlib.entities import Package, PackageCfgFile, PayloadFile
from salt.modules.inspectlib.exceptions import InspectorQueryException, SIException

log = logging.getLogger(__name__)


class SysInfo(object):
    """
    System information.
    """

    def __init__(self, systype):
        if systype.lower() == "solaris":
            raise SIException("Platform {0} not (yet) supported.".format(systype))

    def _grain(self, grain):
        """
        An alias for grains getter.
        """
        return __grains__.get(grain, "N/A")

    def _get_disk_size(self, device):
        """
        Get a size of a disk.
        """
        out = __salt__["cmd.run_all"]("df {0}".format(device))
        if out["retcode"]:
            msg = "Disk size info error: {0}".format(out["stderr"])
            log.error(msg)
            raise SIException(msg)

        devpath, blocks, used, available, used_p, mountpoint = [
            elm for elm in out["stdout"].split(os.linesep)[-1].split(" ") if elm
        ]
        return {
            "device": devpath,
            "blocks": blocks,
            "used": used,
            "available": available,
            "used (%)": used_p,
            "mounted": mountpoint,
        }

    def _get_fs(self):
        """
        Get available file systems and their types.
        """

        data = dict()
        for dev, dev_data in salt.utils.fsutils._blkid().items():
            dev = self._get_disk_size(dev)
            device = dev.pop("device")
            dev["type"] = dev_data["type"]
            data[device] = dev

        return data

    def _get_mounts(self):
        """
        Get mounted FS on the system.
        """
        return salt.utils.fsutils._get_mounts()

    def _get_cpu(self):
        """
        Get available CPU information.
        """
        # CPU data in grains is OK-ish, but lscpu is still better in this case
        out = __salt__["cmd.run_all"]("lscpu")
        salt.utils.fsutils._verify_run(out)
        data = dict()
        for descr, value in [
            elm.split(":", 1) for elm in out["stdout"].split(os.linesep)
        ]:
            data[descr.strip()] = value.strip()

        return data

    def _get_mem(self):
        """
        Get memory.
        """
        out = __salt__["cmd.run_all"]("vmstat -s")
        if out["retcode"]:
            raise SIException("Memory info error: {0}".format(out["stderr"]))

        ret = dict()
        for line in out["stdout"].split(os.linesep):
            line = line.strip()
            if not line:
                continue
            size, descr = line.split(" ", 1)
            if descr.startswith("K "):
                descr = descr[2:]
                size = size + "K"
            ret[descr] = size
        return ret

    def _get_network(self):
        """
        Get network configuration.
        """
        data = dict()
        data["interfaces"] = salt.utils.network.interfaces()
        data["subnets"] = salt.utils.network.subnets()

        return data

    def _get_os(self):
        """
        Get operating system summary
        """
        return {
            "name": self._grain("os"),
            "family": self._grain("os_family"),
            "arch": self._grain("osarch"),
            "release": self._grain("osrelease"),
        }


class Query(EnvLoader):
    """
    Query the system.
    This class is actually puts all Salt features together,
    so there would be no need to pick it from various places.
    """

    # Configuration: config files
    # Identity: users/groups
    # Software: packages, patterns, repositories
    # Services
    # System: distro, RAM etc
    # Changes: all files that are managed and were changed from the original
    # all: include all scopes (scary!)
    # payload: files that are not managed

    SCOPES = [
        "changes",
        "configuration",
        "identity",
        "system",
        "software",
        "services",
        "payload",
        "all",
    ]

    def __init__(self, scope, cachedir=None):
        """
        Constructor.

        :param scope:
        :return:
        """
        if scope and scope not in self.SCOPES:
            raise InspectorQueryException(
                "Unknown scope: {0}. Must be one of: {1}".format(
                    repr(scope), ", ".join(self.SCOPES)
                )
            )
        elif not scope:
            raise InspectorQueryException(
                "Scope cannot be empty. Must be one of: {0}".format(
                    ", ".join(self.SCOPES)
                )
            )
        EnvLoader.__init__(self, cachedir=cachedir)
        self.scope = "_" + scope
        self.local_identity = dict()

    def __call__(self, *args, **kwargs):
        """
        Call the query with the defined scope.

        :param args:
        :param kwargs:
        :return:
        """

        return getattr(self, self.scope)(*args, **kwargs)

    def _changes(self, *args, **kwargs):
        """
        Returns all diffs to the configuration files.
        """
        raise Exception("Not yet implemented")

    def _configuration(self, *args, **kwargs):
        """
        Return configuration files.
        """

        data = dict()
        self.db.open()
        for pkg in self.db.get(Package):
            configs = list()
            for pkg_cfg in self.db.get(PackageCfgFile, eq={"pkgid": pkg.id}):
                configs.append(pkg_cfg.path)
            data[pkg.name] = configs

        if not data:
            raise InspectorQueryException("No inspected configuration yet available.")

        return data

    def _get_local_users(self, disabled=None):
        """
        Return all known local accounts to the system.
        """
        users = dict()
        path = "/etc/passwd"
        with salt.utils.files.fopen(path, "r") as fp_:
            for line in fp_:
                line = line.strip()
                if ":" not in line:
                    continue
                name, password, uid, gid, gecos, directory, shell = line.split(":")
                active = not (password == "*" or password.startswith("!"))
                if (
                    (disabled is False and active)
                    or (disabled is True and not active)
                    or disabled is None
                ):
                    users[name] = {
                        "uid": uid,
                        "git": gid,
                        "info": gecos,
                        "home": directory,
                        "shell": shell,
                        "disabled": not active,
                    }

        return users

    def _get_local_groups(self):
        """
        Return all known local groups to the system.
        """
        groups = dict()
        path = "/etc/group"
        with salt.utils.files.fopen(path, "r") as fp_:
            for line in fp_:
                line = line.strip()
                if ":" not in line:
                    continue
                name, password, gid, users = line.split(":")
                groups[name] = {
                    "gid": gid,
                }

                if users:
                    groups[name]["users"] = users.split(",")

        return groups

    def _get_external_accounts(self, locals):
        """
        Return all known accounts, excluding local accounts.
        """
        users = dict()
        out = __salt__["cmd.run_all"]("passwd -S -a")
        if out["retcode"]:
            # System does not supports all accounts descriptions, just skipping.
            return users
        status = {
            "L": "Locked",
            "NP": "No password",
            "P": "Usable password",
            "LK": "Locked",
        }
        for data in [
            elm.strip().split(" ")
            for elm in out["stdout"].split(os.linesep)
            if elm.strip()
        ]:
            if len(data) < 2:
                continue
            name, login = data[:2]
            if name not in locals:
                users[name] = {"login": login, "status": status.get(login, "N/A")}

        return users

    def _identity(self, *args, **kwargs):
        """
        Local users and groups.

        accounts
            Can be either 'local', 'remote' or 'all' (equal to "local,remote").
            Remote accounts cannot be resolved on all systems, but only
            those, which supports 'passwd -S -a'.

        disabled
            True (or False, default) to return only disabled accounts.
        """
        LOCAL = "local accounts"
        EXT = "external accounts"

        data = dict()
        data[LOCAL] = self._get_local_users(disabled=kwargs.get("disabled"))
        data[EXT] = self._get_external_accounts(data[LOCAL].keys()) or "N/A"
        data["local groups"] = self._get_local_groups()

        return data

    def _system(self, *args, **kwargs):
        """
        This basically calls grains items and picks out only
        necessary information in a certain structure.

        :param args:
        :param kwargs:
        :return:
        """
        sysinfo = SysInfo(__grains__.get("kernel"))

        data = dict()
        data["cpu"] = sysinfo._get_cpu()
        data["disks"] = sysinfo._get_fs()
        data["mounts"] = sysinfo._get_mounts()
        data["memory"] = sysinfo._get_mem()
        data["network"] = sysinfo._get_network()
        data["os"] = sysinfo._get_os()

        return data

    def _software(self, *args, **kwargs):
        """
        Return installed software.
        """
        data = dict()
        if "exclude" in kwargs:
            excludes = kwargs["exclude"].split(",")
        else:
            excludes = list()

        os_family = __grains__.get("os_family").lower()

        # Get locks
        if os_family == "suse":
            LOCKS = "pkg.list_locks"
            if "products" not in excludes:
                products = __salt__["pkg.list_products"]()
                if products:
                    data["products"] = products
        elif os_family == "redhat":
            LOCKS = "pkg.get_locked_packages"
        else:
            LOCKS = None

        if LOCKS and "locks" not in excludes:
            locks = __salt__[LOCKS]()
            if locks:
                data["locks"] = locks

        # Get patterns
        if os_family == "suse":
            PATTERNS = "pkg.list_installed_patterns"
        elif os_family == "redhat":
            PATTERNS = "pkg.group_list"
        else:
            PATTERNS = None

        if PATTERNS and "patterns" not in excludes:
            patterns = __salt__[PATTERNS]()
            if patterns:
                data["patterns"] = patterns

        # Get packages
        if "packages" not in excludes:
            data["packages"] = __salt__["pkg.list_pkgs"]()

        # Get repositories
        if "repositories" not in excludes:
            repos = __salt__["pkg.list_repos"]()
            if repos:
                data["repositories"] = repos

        return data

    def _services(self, *args, **kwargs):
        """
        Get list of enabled and disabled services on the particular system.
        """
        return {
            "enabled": __salt__["service.get_enabled"](),
            "disabled": __salt__["service.get_disabled"](),
        }

    def _id_resolv(self, iid, named=True, uid=True):
        """
        Resolve local users and groups.

        :param iid:
        :param named:
        :param uid:
        :return:
        """

        if not self.local_identity:
            self.local_identity["users"] = self._get_local_users()
            self.local_identity["groups"] = self._get_local_groups()

        if not named:
            return iid

        for name, meta in self.local_identity[uid and "users" or "groups"].items():
            if (uid and int(meta.get("uid", -1)) == iid) or (
                not uid and int(meta.get("gid", -1)) == iid
            ):
                return name

        return iid

    def _payload(self, *args, **kwargs):
        """
        Find all unmanaged files. Returns maximum 1000 values.

        Parameters:

        * **filter**: Include only results which path starts from the filter string.
        * **time**: Display time in Unix ticks or format according to the configured TZ (default)
                    Values: ticks, tz (default)
        * **size**: Format size. Values: B, KB, MB, GB
        * **owners**: Resolve UID/GID to an actual names or leave them numeric (default).
                      Values: name (default), id
        * **type**: Comma-separated type of included payload: dir (or directory), link and/or file.
        * **brief**: Return just a list of matches, if True. Default: False
        * **offset**: Offset of the files
        * **max**: Maximum returned values. Default 1000.

        Options:

        * **total**: Return a total amount of found payload files
        """

        def _size_format(size, fmt):
            if fmt is None:
                return size

            fmt = fmt.lower()
            if fmt == "b":
                return "{0} Bytes".format(size)
            elif fmt == "kb":
                return "{0} Kb".format(round((float(size) / 0x400), 2))
            elif fmt == "mb":
                return "{0} Mb".format(round((float(size) / 0x400 / 0x400), 2))
            elif fmt == "gb":
                return "{0} Gb".format(round((float(size) / 0x400 / 0x400 / 0x400), 2))

        filter = kwargs.get("filter")
        offset = kwargs.get("offset", 0)

        timeformat = kwargs.get("time", "tz")
        if timeformat not in ["ticks", "tz"]:
            raise InspectorQueryException(
                'Unknown "{0}" value for parameter "time"'.format(timeformat)
            )
        tfmt = (
            lambda param: timeformat == "tz"
            and time.strftime("%b %d %Y %H:%M:%S", time.gmtime(param))
            or int(param)
        )

        size_fmt = kwargs.get("size")
        if size_fmt is not None and size_fmt.lower() not in ["b", "kb", "mb", "gb"]:
            raise InspectorQueryException(
                'Unknown "{0}" value for parameter "size". '
                "Should be either B, Kb, Mb or Gb".format(timeformat)
            )

        owners = kwargs.get("owners", "id")
        if owners not in ["name", "id"]:
            raise InspectorQueryException(
                'Unknown "{0}" value for parameter "owners". '
                "Should be either name or id (default)".format(owners)
            )

        incl_type = [prm for prm in kwargs.get("type", "").lower().split(",") if prm]
        if not incl_type:
            incl_type.append("file")

        for i_type in incl_type:
            if i_type not in ["directory", "dir", "d", "file", "f", "link", "l"]:
                raise InspectorQueryException(
                    'Unknown "{0}" values for parameter "type". '
                    "Should be comma separated one or more of "
                    "dir, file and/or link.".format(", ".join(incl_type))
                )
        self.db.open()

        if "total" in args:
            return {"total": len(self.db.get(PayloadFile))}

        brief = kwargs.get("brief")
        pld_files = list() if brief else dict()
        for pld_data in self.db.get(PayloadFile)[
            offset : offset + kwargs.get("max", 1000)
        ]:
            if brief:
                pld_files.append(pld_data.path)
            else:
                pld_files[pld_data.path] = {
                    "uid": self._id_resolv(pld_data.uid, named=(owners == "id")),
                    "gid": self._id_resolv(
                        pld_data.gid, named=(owners == "id"), uid=False
                    ),
                    "size": _size_format(pld_data.p_size, fmt=size_fmt),
                    "mode": oct(pld_data.mode),
                    "accessed": tfmt(pld_data.atime),
                    "modified": tfmt(pld_data.mtime),
                    "created": tfmt(pld_data.ctime),
                }

        return pld_files

    def _all(self, *args, **kwargs):
        """
        Return all the summary of the particular system.
        """
        data = dict()
        data["software"] = self._software(**kwargs)
        data["system"] = self._system(**kwargs)
        data["services"] = self._services(**kwargs)
        try:
            data["configuration"] = self._configuration(**kwargs)
        except InspectorQueryException as ex:
            data["configuration"] = "N/A"
            log.error(ex)
        data["payload"] = self._payload(**kwargs) or "N/A"

        return data



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.pecl as pecl

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PeclTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.pecl
    """

    def setup_loader_modules(self):
        return {pecl: {}}

    # 'installed' function tests: 1

    def test_installed(self):
        """
        Test to make sure that a pecl extension is installed.
        """
        name = "mongo"
        ver = "1.0.1"

        ret = {"name": name, "result": None, "comment": "", "changes": {}}

        mock_lst = MagicMock(return_value={name: "stable"})
        mock_t = MagicMock(return_value=True)
        with patch.dict(pecl.__salt__, {"pecl.list": mock_lst, "pecl.install": mock_t}):
            comt = "Pecl extension {0} is already installed.".format(name)
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(pecl.installed(name), ret)

            with patch.dict(pecl.__opts__, {"test": True}):
                comt = "Pecl extension mongo-1.0.1 would have been installed"
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(pecl.installed(name, version=ver), ret)

            with patch.dict(pecl.__opts__, {"test": False}):
                comt = "Pecl extension mongo-1.0.1 was successfully installed"
                ret.update(
                    {
                        "comment": comt,
                        "result": True,
                        "changes": {"mongo-1.0.1": "Installed"},
                    }
                )
                self.assertDictEqual(pecl.installed(name, version=ver), ret)

    # 'removed' function tests: 1

    def test_removed(self):
        """
        Test to make sure that a pecl extension is not installed.
        """
        name = "mongo"

        ret = {"name": name, "result": None, "comment": "", "changes": {}}

        mock_lst = MagicMock(side_effect=[{}, {name: "stable"}, {name: "stable"}])
        mock_t = MagicMock(return_value=True)
        with patch.dict(
            pecl.__salt__, {"pecl.list": mock_lst, "pecl.uninstall": mock_t}
        ):
            comt = "Pecl extension {0} is not installed.".format(name)
            ret.update({"comment": comt, "result": True})
            self.assertDictEqual(pecl.removed(name), ret)

            with patch.dict(pecl.__opts__, {"test": True}):
                comt = "Pecl extension mongo would have been removed"
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(pecl.removed(name), ret)

            with patch.dict(pecl.__opts__, {"test": False}):
                comt = "Pecl extension mongo was successfully removed."
                ret.update(
                    {"comment": comt, "result": True, "changes": {name: "Removed"}}
                )
                self.assertDictEqual(pecl.removed(name), ret)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Oleg Lipovchenko <oleg.lipovchenko@gmail.com>
"""

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import salt.loader
import salt.matchers.compound_match as compound_match
import salt.matchers.glob_match as glob_match
import salt.matchers.list_match as list_match
import salt.matchers.pcre_match as pcre_match
import salt.modules.match as match

# Import Salt Libs
from salt.exceptions import SaltException

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

MATCHERS_DICT = {
    "compound_match.match": compound_match.match,
    "glob_match.match": glob_match.match,
    "list_match.match": list_match.match,
    "pcre_match.match": pcre_match.match,
}

# the name of the minion to be used for tests
MINION_ID = "bar03"


@patch("salt.loader.matchers", MagicMock(return_value=MATCHERS_DICT))
class MatchTestCase(TestCase, LoaderModuleMockMixin):
    """
    This class contains a set of functions that test salt.modules.match.
    """

    def setup_loader_modules(self):
        return {
            match: {"__opts__": {"extension_modules": "", "id": MINION_ID}},
            compound_match: {"__opts__": {"id": MINION_ID}},
            glob_match: {"__opts__": {"id": MINION_ID}},
            list_match: {"__opts__": {"id": MINION_ID}},
        }

    def test_compound_with_minion_id(self):
        """
        Make sure that when a minion_id IS past, that it is contained in opts
        """
        mock_compound_match = MagicMock()
        target = "bar04"
        new_minion_id = "new_minion_id"

        with patch.object(
            salt.loader,
            "matchers",
            return_value={"compound_match.match": mock_compound_match},
        ) as matchers:
            match.compound(target, minion_id=new_minion_id)

            # The matcher should get called with MINION_ID
            matchers.assert_called_once()
            matchers_opts = matchers.call_args[0][0]
            self.assertEqual(matchers_opts.get("id"), new_minion_id)

            # The compound matcher should not get MINION_ID, no opts should be passed
            mock_compound_match.assert_called_once_with(target)

    def test_compound(self):
        """
        Test issue #55149
        """
        mock_compound_match = MagicMock()
        target = "bar04"

        with patch.object(
            salt.loader,
            "matchers",
            return_value={"compound_match.match": mock_compound_match},
        ) as matchers:
            match.compound(target)

            # The matcher should get called with MINION_ID
            matchers.assert_called_once()
            self.assertEqual(len(matchers.call_args[0]), 1)
            self.assertEqual(matchers.call_args[0][0].get("id"), MINION_ID)

            # The compound matcher should not get MINION_ID, no opts should be passed
            mock_compound_match.assert_called_once_with(target)

    def test_filter_by(self):
        """
        Tests if filter_by returns the correct dictionary.
        """
        lookup = {
            "foo*": {"key1": "fooval1", "key2": "fooval2"},
            "bar*": {"key1": "barval1", "key2": "barval2"},
        }
        result = {"key1": "barval1", "key2": "barval2"}

        self.assertDictEqual(match.filter_by(lookup), result)

    def test_watch_for_opts_mismatch_glob_match(self):
        """
        Tests for situations where the glob matcher might reference __opts__ directly
        instead of the local opts variable.

        When metaproxies/proxy minions are in use, matchers get called with a different `opts`
        dictionary.  Inside the matchers we check to see if `opts` was passed
        and use it instead of `__opts__`.  If sometime in the future we update the matchers
        and use `__opts__` directly this breaks proxy matching.
        """
        self.assertTrue(glob_match.match("bar03"))
        self.assertTrue(glob_match.match("rest03", {"id": "rest03"}))
        self.assertFalse(glob_match.match("rest03"))

    def test_watch_for_opts_mismatch_list_match(self):
        """
        Tests for situations where the list matcher might reference __opts__ directly
        instead of the local opts variable

        When metaproxies/proxy minions are in use, matchers get called with a different `opts`
        dictionary.  Inside the matchers we check to see if `opts` was passed
        and use it instead of `__opts__`.  If sometime in the future we update the matchers
        and use `__opts__` directly this breaks proxy matching.
        """
        self.assertTrue(list_match.match("bar03"))
        self.assertTrue(list_match.match("rest03", {"id": "rest03"}))
        self.assertFalse(list_match.match("rest03"))

    def test_watch_for_opts_mismatch_compound_match(self):
        """
        Tests for situations where the compound matcher might reference __opts__ directly
        instead of the local opts variable

        When metaproxies/proxy minions are in use, matchers get called with a different `opts`
        dictionary.  Inside the matchers we check to see if `opts` was passed
        and use it instead of `__opts__`.  If sometime in the future we update the matchers
        and use `__opts__` directly this breaks proxy matching.
        """
        self.assertTrue(compound_match.match("L@bar03"))
        self.assertTrue(compound_match.match("L@rest03", {"id": "rest03"}))
        self.assertFalse(compound_match.match("L@rest03"))

    def test_filter_by_merge(self):
        """
        Tests if filter_by returns a dictionary merged with another dictionary.
        """
        lookup = {
            "foo*": {"key1": "fooval1", "key2": "fooval2"},
            "bar*": {"key1": "barval1", "key2": "barval2"},
        }
        mdict = {"key1": "mergeval1"}
        result = {"key1": "mergeval1", "key2": "barval2"}

        self.assertDictEqual(match.filter_by(lookup, merge=mdict), result)

    def test_filter_by_merge_lists_rep(self):
        """
        Tests if filter_by merges list values by replacing the original list
        values with the merged list values.
        """
        lookup = {"foo*": {"list_key": []}, "bar*": {"list_key": ["val1", "val2"]}}

        mdict = {"list_key": ["val3", "val4"]}

        # list replacement specified by the merge_lists=False option
        result = {"list_key": ["val3", "val4"]}

        self.assertDictEqual(
            match.filter_by(lookup, merge=mdict, merge_lists=False), result
        )

    def test_filter_by_merge_lists_agg(self):
        """
        Tests if filter_by merges list values by aggregating them.
        """
        lookup = {"foo*": {"list_key": []}, "bar*": {"list_key": ["val1", "val2"]}}

        mdict = {"list_key": ["val3", "val4"]}

        # list aggregation specified by the merge_lists=True option
        result = {"list_key": ["val1", "val2", "val3", "val4"]}

        self.assertDictEqual(
            match.filter_by(lookup, merge=mdict, merge_lists=True), result
        )

    def test_filter_by_merge_with_none(self):
        """
        Tests if filter_by merges a None object with a merge dictionary.
        """
        lookup = {"foo*": {"key1": "fooval1", "key2": "fooval2"}, "bar*": None}

        # mdict should also be the returned dictionary
        # since a merge is done with None
        mdict = {"key1": "mergeval1"}

        self.assertDictEqual(match.filter_by(lookup, merge=mdict), mdict)

    def test_filter_by_merge_fail(self):
        """
        Tests for an exception if a merge is done without a dictionary.
        """
        lookup = {
            "foo*": {"key1": "fooval1", "key2": "fooval2"},
            "bar*": {"key1": "barval1", "key2": "barval2"},
        }
        mdict = "notadict"

        self.assertRaises(SaltException, match.filter_by, lookup, merge=mdict)

    def test_glob_match_different_minon_id(self):
        """
        Tests for situations where the glob matcher is called with a different
        minion_id than what is found in __opts__
        """
        # not passing minion_id, should return False
        self.assertFalse(match.glob("bar04"))

        # passing minion_id, should return True
        self.assertTrue(match.glob("bar04", "bar04"))

    def test_pcre_match_different_minion_id(self):
        """
        Tests for situations where the glob matcher is called with a different
        minion_id than what is found in __opts__
        """
        # not passing minion_id, should return False
        self.assertFalse(match.pcre("bar.*04"))

        # passing minion_id, should return True
        self.assertTrue(match.pcre("bar.*04", "bar04"))

    def test_list_match_different_minion_id(self):
        """
        Tests for situations where the list matcher is called with a different
        minion_id than what is found in __opts__
        """
        # not passing minion_id, should return False
        self.assertFalse(match.list_("bar02,bar04"))

        # passing minion_id, should return True
        self.assertTrue(match.list_("bar02,bar04", "bar04"))



# -*- coding: utf-8 -*-
"""
Unit tests for the docker module
"""
from __future__ import absolute_import, print_function, unicode_literals

import logging

import salt.config
import salt.loader
import salt.modules.dockermod as docker_mod
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.ext.six.moves import range
from tests.support.helpers import slowTest
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, call, patch
from tests.support.unit import TestCase, skipIf

log = logging.getLogger(__name__)


def _docker_py_version():
    try:
        if docker_mod.HAS_DOCKER_PY:
            return docker_mod.docker.version_info
    except AttributeError:
        pass
    return (0,)


class DockerUnitTestCase(TestCase, LoaderModuleMockMixin):
    def fake_run(self, *args, **kwargs):
        print(args, kwargs)
        return "{}"

    def setup_loader_modules(self):
        return {
            docker_mod: {
                "__utils__": {
                    "state.get_sls_opts": MagicMock(
                        return_value={
                            "pillarenv": MagicMock(),
                            "pillar": {},
                            "grains": {},
                        }
                    ),
                    "args.clean_kwargs": lambda **x: x,
                },
                "__salt__": {
                    "config.option": MagicMock(return_value=None),
                    "cmd.run": self.fake_run,
                },
                "__opts__": {"id": "dockermod-unit-test"},
            },
        }

    def test_trans_tar_should_have_grains_in_sls_opts_including_pillar_override(self):
        container_name = "fnord"
        expected_grains = {
            "roscivs": "bottia",
            "fnord": "dronf",
            "salt": "NaCl",
        }
        expected_pillars = {
            "this": {"is": {"my": {"pillar": "data"}}},
        }
        extra_pillar_data = {"some": "extras"}
        fake_trans_tar = MagicMock(return_value=b"hi")
        patch_trans_tar = patch(
            "salt.modules.dockermod._prepare_trans_tar", fake_trans_tar,
        )
        patch_call = patch(
            "salt.modules.dockermod.call", MagicMock(return_value=expected_grains),
        )
        fake_get_pillar = MagicMock()
        fake_get_pillar.compile_pillar.return_value = expected_pillars
        patch_pillar = patch(
            "salt.modules.dockermod.salt.pillar.get_pillar",
            MagicMock(return_value=fake_get_pillar),
        )
        patch_run_all = patch(
            "salt.modules.dockermod.run_all",
            MagicMock(return_value={"retcode": 1, "stderr": "early exit test"}),
        )
        with patch_trans_tar, patch_call, patch_pillar, patch_run_all:
            docker_mod.sls(container_name, pillar=extra_pillar_data)
            # TODO: It would be fine if we could make this test require less magic numbers -W. Werner, 2019-08-27
            actual_sls_opts = fake_trans_tar.call_args[0][1]
            self.assertDictContainsSubset(
                expected_grains,
                actual_sls_opts["grains"],
                "Docker container grains not provided to thin client creation",
            )
            expected_pillars.update(extra_pillar_data)
            self.assertDictContainsSubset(
                expected_pillars,
                actual_sls_opts["pillar"],
                "Docker container pillar not provided to thin client creation",
            )


@skipIf(
    docker_mod.HAS_DOCKER_PY is False,
    "docker-py must be installed to run these tests. Skipping.",
)
class DockerTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate docker module
    """

    def setup_loader_modules(self):
        utils = salt.loader.utils(
            salt.config.DEFAULT_MINION_OPTS.copy(),
            whitelist=[
                "args",
                "docker",
                "json",
                "state",
                "thin",
                "systemd",
                "path",
                "platform",
            ],
        )
        return {
            docker_mod: {
                "__context__": {"docker.docker_version": ""},
                "__utils__": utils,
            }
        }

    try:
        docker_version = docker_mod.docker.version_info
    except AttributeError:
        docker_version = (0,)

    def setUp(self):
        """
        Ensure we aren't persisting context dunders between tests
        """
        docker_mod.__context__.pop("docker.client", None)

    def test_failed_login(self):
        """
        Check that when docker.login failed a retcode other then 0
        is part of the return.
        """
        client = Mock()
        get_client_mock = MagicMock(return_value=client)
        ref_out = {"stdout": "", "stderr": "login failed", "retcode": 1}
        with patch.dict(
            docker_mod.__pillar__,
            {
                "docker-registries": {
                    "portus.example.com:5000": {
                        "username": "admin",
                        "password": "linux12345",
                        "email": "tux@example.com",
                    }
                }
            },
        ):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                dunder_salt = {
                    "cmd.run_all": MagicMock(return_value=ref_out),
                    "config.get": MagicMock(return_value={}),
                    "config.option": MagicMock(return_value={}),
                }
                with patch.dict(docker_mod.__salt__, dunder_salt):
                    ret = docker_mod.login("portus.example.com:5000")
                    self.assertIn("retcode", ret)
                    self.assertNotEqual(ret["retcode"], 0)

    def test_logout_calls_docker_cli_logout_single(self):
        client = Mock()
        get_client_mock = MagicMock(return_value=client)
        ref_out = {"stdout": "", "stderr": "", "retcode": 0}
        registry_auth_data = {
            "portus.example.com:5000": {
                "username": "admin",
                "password": "linux12345",
                "email": "tux@example.com",
            }
        }
        docker_mock = MagicMock(return_value=ref_out)
        with patch.object(docker_mod, "_get_client", get_client_mock):
            dunder_salt = {
                "config.get": MagicMock(return_value=registry_auth_data),
                "cmd.run_all": docker_mock,
                "config.option": MagicMock(return_value={}),
            }
            with patch.dict(docker_mod.__salt__, dunder_salt):
                ret = docker_mod.logout("portus.example.com:5000")
                assert "retcode" in ret
                assert ret["retcode"] == 0
                docker_mock.assert_called_with(
                    ["docker", "logout", "portus.example.com:5000"],
                    python_shell=False,
                    output_loglevel="quiet",
                )

    def test_logout_calls_docker_cli_logout_all(self):
        client = Mock()
        get_client_mock = MagicMock(return_value=client)
        ref_out = {"stdout": "", "stderr": "", "retcode": 0}
        registry_auth_data = {
            "portus.example.com:5000": {
                "username": "admin",
                "password": "linux12345",
                "email": "tux@example.com",
            },
            "portus2.example.com:5000": {
                "username": "admin",
                "password": "linux12345",
                "email": "tux@example.com",
            },
        }

        docker_mock = MagicMock(return_value=ref_out)
        with patch.object(docker_mod, "_get_client", get_client_mock):
            dunder_salt = {
                "config.get": MagicMock(return_value=registry_auth_data),
                "cmd.run_all": docker_mock,
                "config.option": MagicMock(return_value={}),
            }
            with patch.dict(docker_mod.__salt__, dunder_salt):
                ret = docker_mod.logout()
                assert "retcode" in ret
                assert ret["retcode"] == 0
                assert docker_mock.call_count == 2

    def test_ps_with_host_true(self):
        """
        Check that docker.ps called with host is ``True``,
        include resutlt of ``network.interfaces`` command in returned result.
        """
        client = Mock()
        client.containers = MagicMock(return_value=[])
        get_client_mock = MagicMock(return_value=client)
        network_interfaces = Mock(return_value={"mocked": None})

        with patch.dict(
            docker_mod.__salt__, {"network.interfaces": network_interfaces}
        ):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                ret = docker_mod.ps_(host=True)
                self.assertEqual(ret, {"host": {"interfaces": {"mocked": None}}})

    def test_ps_with_filters(self):
        """
        Check that docker.ps accept filters parameter.
        """
        client = Mock()
        client.containers = MagicMock(return_value=[])
        get_client_mock = MagicMock(return_value=client)

        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.ps_(filters={"label": "KEY"})
            client.containers.assert_called_once_with(
                all=True, filters={"label": "KEY"}
            )

    @slowTest
    def test_check_mine_cache_is_refreshed_on_container_change_event(self):
        """
        Every command that might modify docker containers state.
        Should trig an update on ``mine.send``
        """
        with patch.object(docker_mod, "_get_exec_driver"):
            client_args_mock = MagicMock(
                return_value={
                    "create_container": [
                        "image",
                        "command",
                        "hostname",
                        "user",
                        "detach",
                        "stdin_open",
                        "tty",
                        "ports",
                        "environment",
                        "volumes",
                        "network_disabled",
                        "name",
                        "entrypoint",
                        "working_dir",
                        "domainname",
                        "cpuset",
                        "host_config",
                        "mac_address",
                        "labels",
                        "volume_driver",
                        "stop_signal",
                        "networking_config",
                        "healthcheck",
                        "stop_timeout",
                    ],
                    "host_config": [
                        "binds",
                        "port_bindings",
                        "lxc_conf",
                        "publish_all_ports",
                        "links",
                        "privileged",
                        "dns",
                        "dns_search",
                        "volumes_from",
                        "network_mode",
                        "restart_policy",
                        "cap_add",
                        "cap_drop",
                        "devices",
                        "extra_hosts",
                        "read_only",
                        "pid_mode",
                        "ipc_mode",
                        "security_opt",
                        "ulimits",
                        "log_config",
                        "mem_limit",
                        "memswap_limit",
                        "mem_reservation",
                        "kernel_memory",
                        "mem_swappiness",
                        "cgroup_parent",
                        "group_add",
                        "cpu_quota",
                        "cpu_period",
                        "blkio_weight",
                        "blkio_weight_device",
                        "device_read_bps",
                        "device_write_bps",
                        "device_read_iops",
                        "device_write_iops",
                        "oom_kill_disable",
                        "shm_size",
                        "sysctls",
                        "tmpfs",
                        "oom_score_adj",
                        "dns_opt",
                        "cpu_shares",
                        "cpuset_cpus",
                        "userns_mode",
                        "pids_limit",
                        "isolation",
                        "auto_remove",
                        "storage_opt",
                    ],
                    "networking_config": [
                        "aliases",
                        "links",
                        "ipv4_address",
                        "ipv6_address",
                        "link_local_ips",
                    ],
                }
            )

            for command_name, args in (
                ("create", ()),
                ("rm_", ()),
                ("kill", ()),
                ("pause", ()),
                ("signal_", ("KILL",)),
                ("start_", ()),
                ("stop", ()),
                ("unpause", ()),
                ("_run", ("command",)),
                ("_script", ("command",)),
            ):
                mine_send = Mock()
                command = getattr(docker_mod, command_name)
                client = MagicMock()
                client.api_version = "1.12"
                with patch.dict(
                    docker_mod.__salt__,
                    {
                        "mine.send": mine_send,
                        "container_resource.run": MagicMock(),
                        "config.get": MagicMock(return_value=True),
                        "config.option": MagicMock(return_value=True),
                        "cp.cache_file": MagicMock(return_value=False),
                    },
                ):
                    with patch.dict(
                        docker_mod.__utils__,
                        {"docker.get_client_args": client_args_mock},
                    ):
                        with patch.object(docker_mod, "_get_client", client):
                            command("container", *args)
                try:
                    mine_send.assert_called_with(
                        "docker.ps", verbose=True, all=True, host=True
                    )
                except AssertionError as exc:
                    raise Exception(
                        "command '{0}' did not call docker.ps with expected "
                        "arguments: {1}".format(command_name, exc)
                    )

    def test_update_mine(self):
        """
        Test the docker.update_mine config option
        """

        def config_get_disabled(val, default):
            return {
                "base_url": docker_mod.NOTSET,
                "version": docker_mod.NOTSET,
                "docker.url": docker_mod.NOTSET,
                "docker.version": docker_mod.NOTSET,
                "docker.machine": docker_mod.NOTSET,
                "docker.update_mine": False,
            }[val]

        def config_get_enabled(val, default):
            return {
                "base_url": docker_mod.NOTSET,
                "version": docker_mod.NOTSET,
                "docker.url": docker_mod.NOTSET,
                "docker.version": docker_mod.NOTSET,
                "docker.machine": docker_mod.NOTSET,
                "docker.update_mine": True,
            }[val]

        mine_mock = Mock()
        dunder_salt = {
            "config.get": MagicMock(side_effect=config_get_disabled),
            "config.option": MagicMock(return_value=False),
            "mine.send": mine_mock,
        }
        with patch.dict(docker_mod.__salt__, dunder_salt), patch.dict(
            docker_mod.__context__, {"docker.client": Mock()}
        ), patch.object(docker_mod, "state", MagicMock(return_value="stopped")):
            docker_mod.stop("foo", timeout=1)
            mine_mock.assert_not_called()

        with patch.dict(docker_mod.__salt__, dunder_salt), patch.dict(
            docker_mod.__context__, {"docker.client": Mock()}
        ), patch.object(docker_mod, "state", MagicMock(return_value="stopped")):
            dunder_salt["config.get"].side_effect = config_get_enabled
            dunder_salt["config.option"].return_value = True
            docker_mod.stop("foo", timeout=1)
            self.assert_called_once(mine_mock)

    @skipIf(
        _docker_py_version() < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_list_networks(self, *args):
        """
        test list networks.
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        host_config = {}
        client = Mock()
        client.api_version = "1.21"
        client.networks = Mock(
            return_value=[{"Name": "foo", "Id": "01234", "Containers": {}}]
        )
        get_client_mock = MagicMock(return_value=client)
        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.networks(names=["foo"], ids=["01234"])
        client.networks.assert_called_once_with(names=["foo"], ids=["01234"])

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_create_network(self, *args):
        """
        test create network.
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        host_config = {}
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.create_network(
                    "foo",
                    driver="bridge",
                    driver_opts={},
                    gateway="192.168.0.1",
                    ip_range="192.168.0.128/25",
                    subnet="192.168.0.0/24",
                )
        client.create_network.assert_called_once_with(
            "foo",
            driver="bridge",
            options={},
            ipam={
                "Config": [
                    {
                        "Gateway": "192.168.0.1",
                        "IPRange": "192.168.0.128/25",
                        "Subnet": "192.168.0.0/24",
                    }
                ],
                "Driver": "default",
            },
            check_duplicate=True,
        )

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_remove_network(self, *args):
        """
        test remove network.
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        host_config = {}
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.remove_network("foo")
        client.remove_network.assert_called_once_with("foo")

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_inspect_network(self, *args):
        """
        test inspect network.
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        host_config = {}
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.inspect_network("foo")
        client.inspect_network.assert_called_once_with("foo")

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_connect_container_to_network(self, *args):
        """
        test connect_container_to_network
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        host_config = {}
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        context = {"docker.exec_driver": "docker-exec"}

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.dict(docker_mod.__context__, context):
                with patch.object(docker_mod, "_get_client", get_client_mock):
                    docker_mod.connect_container_to_network("container", "foo")
        client.connect_container_to_network.assert_called_once_with("container", "foo")

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_disconnect_container_from_network(self, *args):
        """
        test disconnect_container_from_network
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        host_config = {}
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.disconnect_container_from_network("container", "foo")
        client.disconnect_container_from_network.assert_called_once_with(
            "container", "foo"
        )

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_list_volumes(self, *args):
        """
        test list volumes.
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.volumes(filters={"dangling": [True]},)
        client.volumes.assert_called_once_with(filters={"dangling": [True]},)

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_create_volume(self, *args):
        """
        test create volume.
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.create_volume(
                    "foo", driver="bridge", driver_opts={},
                )
        client.create_volume.assert_called_once_with(
            "foo", driver="bridge", driver_opts={},
        )

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_remove_volume(self, *args):
        """
        test remove volume.
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.remove_volume("foo")
        client.remove_volume.assert_called_once_with("foo")

    @skipIf(
        docker_version < (1, 5, 0),
        "docker module must be installed to run this test or is too old. >=1.5.0",
    )
    def test_inspect_volume(self, *args):
        """
        test inspect volume.
        """
        __salt__ = {
            "config.get": Mock(),
            "mine.send": Mock(),
        }
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.inspect_volume("foo")
        client.inspect_volume.assert_called_once_with("foo")

    def test_wait_success(self):
        client = Mock()
        client.api_version = "1.21"
        client.wait = Mock(return_value=0)
        get_client_mock = MagicMock(return_value=client)

        docker_inspect_container = Mock(
            side_effect=[{"State": {"Running": True}}, {"State": {"Stopped": True}}]
        )
        with patch.object(docker_mod, "inspect_container", docker_inspect_container):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod._clear_context()
                result = docker_mod.wait("foo")
        self.assertEqual(
            result,
            {
                "result": True,
                "exit_status": 0,
                "state": {"new": "stopped", "old": "running"},
            },
        )

    def test_wait_fails_already_stopped(self):
        client = Mock()
        client.api_version = "1.21"
        client.wait = Mock(return_value=0)
        get_client_mock = MagicMock(return_value=client)

        docker_inspect_container = Mock(
            side_effect=[{"State": {"Stopped": True}}, {"State": {"Stopped": True}}]
        )
        with patch.object(docker_mod, "inspect_container", docker_inspect_container):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod._clear_context()
                result = docker_mod.wait("foo")
        self.assertEqual(
            result,
            {
                "result": False,
                "comment": "Container 'foo' already stopped",
                "exit_status": 0,
                "state": {"new": "stopped", "old": "stopped"},
            },
        )

    def test_wait_success_already_stopped(self):
        client = Mock()
        client.api_version = "1.21"
        client.wait = Mock(return_value=0)
        get_client_mock = MagicMock(return_value=client)

        docker_inspect_container = Mock(
            side_effect=[{"State": {"Stopped": True}}, {"State": {"Stopped": True}}]
        )
        with patch.object(docker_mod, "inspect_container", docker_inspect_container):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod._clear_context()
                result = docker_mod.wait("foo", ignore_already_stopped=True)
        self.assertEqual(
            result,
            {
                "result": True,
                "comment": "Container 'foo' already stopped",
                "exit_status": 0,
                "state": {"new": "stopped", "old": "stopped"},
            },
        )

    def test_wait_success_absent_container(self):
        client = Mock()
        client.api_version = "1.21"
        get_client_mock = MagicMock(return_value=client)

        docker_inspect_container = Mock(side_effect=CommandExecutionError)
        with patch.object(docker_mod, "inspect_container", docker_inspect_container):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod._clear_context()
                result = docker_mod.wait("foo", ignore_already_stopped=True)
        self.assertEqual(result, {"result": True, "comment": "Container 'foo' absent"})

    def test_wait_fails_on_exit_status(self):
        client = Mock()
        client.api_version = "1.21"
        client.wait = Mock(return_value=1)
        get_client_mock = MagicMock(return_value=client)

        docker_inspect_container = Mock(
            side_effect=[{"State": {"Running": True}}, {"State": {"Stopped": True}}]
        )
        with patch.object(docker_mod, "inspect_container", docker_inspect_container):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod._clear_context()
                result = docker_mod.wait("foo", fail_on_exit_status=True)
        self.assertEqual(
            result,
            {
                "result": False,
                "exit_status": 1,
                "state": {"new": "stopped", "old": "running"},
            },
        )

    def test_wait_fails_on_exit_status_and_already_stopped(self):
        client = Mock()
        client.api_version = "1.21"
        client.wait = Mock(return_value=1)
        get_client_mock = MagicMock(return_value=client)

        docker_inspect_container = Mock(
            side_effect=[{"State": {"Stopped": True}}, {"State": {"Stopped": True}}]
        )
        with patch.object(docker_mod, "inspect_container", docker_inspect_container):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod._clear_context()
                result = docker_mod.wait(
                    "foo", ignore_already_stopped=True, fail_on_exit_status=True
                )
        self.assertEqual(
            result,
            {
                "result": False,
                "comment": "Container 'foo' already stopped",
                "exit_status": 1,
                "state": {"new": "stopped", "old": "stopped"},
            },
        )

    def test_sls_build(self, *args):
        """
        test build sls image.
        """
        docker_start_mock = MagicMock(return_value={})
        docker_create_mock = MagicMock(return_value={"Id": "ID", "Name": "NAME"})
        docker_stop_mock = MagicMock(
            return_value={"state": {"old": "running", "new": "stopped"}, "result": True}
        )
        docker_rm_mock = MagicMock(return_value={})
        docker_commit_mock = MagicMock(
            return_value={"Id": "ID2", "Image": "foo", "Time_Elapsed": 42}
        )

        docker_sls_mock = MagicMock(
            return_value={
                "file_|-/etc/test.sh_|-/etc/test.sh_|-managed": {
                    "comment": "File /etc/test.sh is in the correct state",
                    "name": "/etc/test.sh",
                    "start_time": "07:04:26.834792",
                    "result": True,
                    "duration": 13.492,
                    "__run_num__": 0,
                    "changes": {},
                },
                "test_|-always-passes_|-foo_|-succeed_without_changes": {
                    "comment": "Success!",
                    "name": "foo",
                    "start_time": "07:04:26.848915",
                    "result": True,
                    "duration": 0.363,
                    "__run_num__": 1,
                    "changes": {},
                },
            }
        )

        ret = None
        with patch.object(docker_mod, "start_", docker_start_mock), patch.object(
            docker_mod, "create", docker_create_mock
        ), patch.object(docker_mod, "stop", docker_stop_mock), patch.object(
            docker_mod, "commit", docker_commit_mock
        ), patch.object(
            docker_mod, "sls", docker_sls_mock
        ), patch.object(
            docker_mod, "rm_", docker_rm_mock
        ):
            ret = docker_mod.sls_build("foo", mods="foo")
        docker_create_mock.assert_called_once_with(
            cmd="sleep infinity", image="opensuse/python", interactive=True, tty=True
        )
        docker_start_mock.assert_called_once_with("ID")
        docker_sls_mock.assert_called_once_with("ID", "foo")
        docker_stop_mock.assert_called_once_with("ID")
        docker_rm_mock.assert_called_once_with("ID")
        docker_commit_mock.assert_called_once_with("ID", "foo", tag="latest")
        self.assertEqual({"Id": "ID2", "Image": "foo", "Time_Elapsed": 42}, ret)

    def test_sls_build_dryrun(self, *args):
        """
        test build sls image in dryrun mode.
        """
        docker_start_mock = MagicMock(return_value={})
        docker_create_mock = MagicMock(return_value={"Id": "ID", "Name": "NAME"})
        docker_stop_mock = MagicMock(
            return_value={"state": {"old": "running", "new": "stopped"}, "result": True}
        )
        docker_rm_mock = MagicMock(return_value={})

        docker_sls_mock = MagicMock(
            return_value={
                "file_|-/etc/test.sh_|-/etc/test.sh_|-managed": {
                    "comment": "File /etc/test.sh is in the correct state",
                    "name": "/etc/test.sh",
                    "start_time": "07:04:26.834792",
                    "result": True,
                    "duration": 13.492,
                    "__run_num__": 0,
                    "changes": {},
                },
                "test_|-always-passes_|-foo_|-succeed_without_changes": {
                    "comment": "Success!",
                    "name": "foo",
                    "start_time": "07:04:26.848915",
                    "result": True,
                    "duration": 0.363,
                    "__run_num__": 1,
                    "changes": {},
                },
            }
        )

        ret = None
        with patch.object(docker_mod, "start_", docker_start_mock), patch.object(
            docker_mod, "create", docker_create_mock
        ), patch.object(docker_mod, "stop", docker_stop_mock), patch.object(
            docker_mod, "rm_", docker_rm_mock
        ), patch.object(
            docker_mod, "sls", docker_sls_mock
        ):
            ret = docker_mod.sls_build("foo", mods="foo", dryrun=True)
        docker_create_mock.assert_called_once_with(
            cmd="sleep infinity", image="opensuse/python", interactive=True, tty=True
        )
        docker_start_mock.assert_called_once_with("ID")
        docker_sls_mock.assert_called_once_with("ID", "foo")
        docker_stop_mock.assert_called_once_with("ID")
        docker_rm_mock.assert_called_once_with("ID")
        self.assertEqual(
            {
                "file_|-/etc/test.sh_|-/etc/test.sh_|-managed": {
                    "comment": "File /etc/test.sh is in the correct state",
                    "name": "/etc/test.sh",
                    "start_time": "07:04:26.834792",
                    "result": True,
                    "duration": 13.492,
                    "__run_num__": 0,
                    "changes": {},
                },
                "test_|-always-passes_|-foo_|-succeed_without_changes": {
                    "comment": "Success!",
                    "name": "foo",
                    "start_time": "07:04:26.848915",
                    "result": True,
                    "duration": 0.363,
                    "__run_num__": 1,
                    "changes": {},
                },
            },
            ret,
        )

    @slowTest
    def test_call_success(self):
        """
        test module calling inside containers
        """
        ret = None
        docker_run_all_mock = MagicMock(
            return_value={
                "retcode": 0,
                "stdout": '{"retcode": 0, "comment": "container cmd"}',
                "stderr": "err",
            }
        )
        docker_copy_to_mock = MagicMock(return_value={"retcode": 0})
        docker_config_mock = MagicMock(return_value="")
        client = Mock()
        client.put_archive = Mock()
        get_client_mock = MagicMock(return_value=client)

        context = {"docker.exec_driver": "docker-exec"}
        salt_dunder = {"config.option": docker_config_mock}

        with patch.object(docker_mod, "run_all", docker_run_all_mock), patch.object(
            docker_mod, "copy_to", docker_copy_to_mock
        ), patch.object(docker_mod, "_get_client", get_client_mock), patch.dict(
            docker_mod.__opts__, {"cachedir": "/tmp"}
        ), patch.dict(
            docker_mod.__salt__, salt_dunder
        ), patch.dict(
            docker_mod.__context__, context
        ):
            # call twice to verify tmp path later
            for i in range(2):
                ret = docker_mod.call("ID", "test.arg", 1, 2, arg1="val1")

        # Check that the directory is different each time
        # [ call(name, [args]), ...
        self.maxDiff = None
        self.assertIn("mkdir", docker_run_all_mock.mock_calls[0][1][1])
        self.assertIn("mkdir", docker_run_all_mock.mock_calls[4][1][1])
        self.assertNotEqual(
            docker_run_all_mock.mock_calls[0][1][1],
            docker_run_all_mock.mock_calls[4][1][1],
        )

        self.assertIn("salt-call", docker_run_all_mock.mock_calls[2][1][1])
        self.assertIn("salt-call", docker_run_all_mock.mock_calls[6][1][1])
        self.assertNotEqual(
            docker_run_all_mock.mock_calls[2][1][1],
            docker_run_all_mock.mock_calls[6][1][1],
        )

        # check thin untar
        self.assertIn("tarfile", docker_run_all_mock.mock_calls[1][1][1])
        self.assertIn("tarfile", docker_run_all_mock.mock_calls[5][1][1])
        self.assertNotEqual(
            docker_run_all_mock.mock_calls[1][1][1],
            docker_run_all_mock.mock_calls[5][1][1],
        )

        # check directory cleanup
        self.assertIn("rm -rf", docker_run_all_mock.mock_calls[3][1][1])
        self.assertIn("rm -rf", docker_run_all_mock.mock_calls[7][1][1])
        self.assertNotEqual(
            docker_run_all_mock.mock_calls[3][1][1],
            docker_run_all_mock.mock_calls[7][1][1],
        )

        self.assertEqual({"retcode": 0, "comment": "container cmd"}, ret)

    def test_images_with_empty_tags(self):
        """
        docker 1.12 reports also images without tags with `null`.
        """
        client = Mock()
        client.api_version = "1.24"
        client.images = Mock(
            return_value=[
                {"Id": "sha256:abcde", "RepoTags": None},
                {"Id": "sha256:abcdef"},
                {"Id": "sha256:abcdefg", "RepoTags": ["image:latest"]},
            ]
        )
        get_client_mock = MagicMock(return_value=client)

        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod._clear_context()
            result = docker_mod.images()
        self.assertEqual(result, {"sha256:abcdefg": {"RepoTags": ["image:latest"]}})

    def test_compare_container_image_id_resolution(self):
        """
        Test comparing two containers when one's inspect output is an ID and
        not formatted in image:tag notation.
        """

        def _inspect_container_effect(id_):
            return {
                "container1": {
                    "Config": {"Image": "realimage:latest"},
                    "HostConfig": {},
                },
                "container2": {"Config": {"Image": "image_id"}, "HostConfig": {}},
            }[id_]

        def _inspect_image_effect(id_):
            return {
                "realimage:latest": {"Id": "image_id"},
                "image_id": {"Id": "image_id"},
            }[id_]

        inspect_container_mock = MagicMock(side_effect=_inspect_container_effect)
        inspect_image_mock = MagicMock(side_effect=_inspect_image_effect)

        with patch.object(docker_mod, "inspect_container", inspect_container_mock):
            with patch.object(docker_mod, "inspect_image", inspect_image_mock):
                ret = docker_mod.compare_containers("container1", "container2")
                self.assertEqual(ret, {})

    def test_compare_container_ulimits_order(self):
        """
        Test comparing two containers when the order of the Ulimits HostConfig
        values are different, but the values are the same.
        """

        def _inspect_container_effect(id_):
            return {
                "container1": {
                    "Config": {},
                    "HostConfig": {
                        "Ulimits": [
                            {"Hard": -1, "Soft": -1, "Name": "core"},
                            {"Hard": 65536, "Soft": 65536, "Name": "nofile"},
                        ]
                    },
                },
                "container2": {
                    "Config": {},
                    "HostConfig": {
                        "Ulimits": [
                            {"Hard": 65536, "Soft": 65536, "Name": "nofile"},
                            {"Hard": -1, "Soft": -1, "Name": "core"},
                        ]
                    },
                },
            }[id_]

        inspect_container_mock = MagicMock(side_effect=_inspect_container_effect)

        with patch.object(docker_mod, "inspect_container", inspect_container_mock):
            # pylint: disable=not-callable
            ret = docker_mod.compare_container("container1", "container2")
            # pylint: enable=not-callable
            self.assertEqual(ret, {})

    def test_compare_container_env_order(self):
        """
        Test comparing two containers when the order of the Env HostConfig
        values are different, but the values are the same.
        """

        def _inspect_container_effect(id_):
            return {
                "container1": {
                    "Config": {},
                    "HostConfig": {"Env": ["FOO=bar", "HELLO=world"]},
                },
                "container2": {
                    "Config": {},
                    "HostConfig": {"Env": ["HELLO=world", "FOO=bar"]},
                },
            }[id_]

        inspect_container_mock = MagicMock(side_effect=_inspect_container_effect)

        with patch.object(docker_mod, "inspect_container", inspect_container_mock):
            # pylint: disable=not-callable
            ret = docker_mod.compare_container("container1", "container2")
            # pylint: enable=not-callable
            self.assertEqual(ret, {})

    def test_resolve_tag(self):
        """
        Test the resolve_tag function. It runs docker.insect_image on the image
        name passed and then looks for the RepoTags key in the result
        """
        id_ = "sha256:6ad733544a6317992a6fac4eb19fe1df577d4dec7529efec28a5bd0edad0fd30"
        tags = ["foo:latest", "foo:bar"]
        mock_tagged = MagicMock(return_value={"Id": id_, "RepoTags": tags})
        mock_untagged = MagicMock(return_value={"Id": id_, "RepoTags": []})
        mock_unexpected = MagicMock(return_value={"Id": id_})
        mock_not_found = MagicMock(side_effect=CommandExecutionError())

        with patch.object(docker_mod, "inspect_image", mock_tagged):
            self.assertEqual(docker_mod.resolve_tag("foo"), tags[0])
            self.assertEqual(docker_mod.resolve_tag("foo", all=True), tags)

        with patch.object(docker_mod, "inspect_image", mock_untagged):
            self.assertEqual(docker_mod.resolve_tag("foo"), id_)
            self.assertEqual(docker_mod.resolve_tag("foo", all=True), [id_])

        with patch.object(docker_mod, "inspect_image", mock_unexpected):
            self.assertEqual(docker_mod.resolve_tag("foo"), id_)
            self.assertEqual(docker_mod.resolve_tag("foo", all=True), [id_])

        with patch.object(docker_mod, "inspect_image", mock_not_found):
            self.assertIs(docker_mod.resolve_tag("foo"), False)
            self.assertIs(docker_mod.resolve_tag("foo", all=True), False)

    def test_prune(self):
        """
        Test the prune function
        """

        def _run(**kwargs):
            side_effect = kwargs.pop("side_effect", None)
            # No arguments passed, we should be pruning everything but volumes
            client = Mock()
            if side_effect is not None:
                client.side_effect = side_effect
            with patch.object(docker_mod, "_client_wrapper", client):
                docker_mod.prune(**kwargs)
            return client

        # Containers only, no filters
        client = _run(containers=True)
        client.assert_called_once_with("prune_containers", filters={})

        # Containers only, with filters
        client = _run(containers=True, until="24h", label="foo,bar=baz")
        client.assert_called_once_with(
            "prune_containers", filters={"until": ["24h"], "label": ["foo", "bar=baz"]}
        )

        # Images only, no filters
        client = _run(images=True)
        client.assert_called_once_with("prune_images", filters={})

        # Images only, with filters
        client = _run(images=True, dangling=True, until="24h", label="foo,bar=baz")
        client.assert_called_once_with(
            "prune_images",
            filters={"dangling": True, "until": ["24h"], "label": ["foo", "bar=baz"]},
        )

        # Networks only, no filters
        client = _run(networks=True)
        client.assert_called_once_with("prune_networks", filters={})

        # Networks only, with filters
        client = _run(networks=True, until="24h", label="foo,bar=baz")
        client.assert_called_once_with(
            "prune_networks", filters={"until": ["24h"], "label": ["foo", "bar=baz"]}
        )

        # Volumes only, no filters
        client = _run(system=False, volumes=True)
        client.assert_called_once_with("prune_volumes", filters={})

        # Volumes only, with filters
        client = _run(system=False, volumes=True, label="foo,bar=baz")
        client.assert_called_once_with(
            "prune_volumes", filters={"label": ["foo", "bar=baz"]}
        )

        # Containers and images, no filters
        client = _run(containers=True, images=True)
        self.assertEqual(
            client.call_args_list,
            [call("prune_containers", filters={}), call("prune_images", filters={})],
        )

        # Containers and images, with filters
        client = _run(containers=True, images=True, until="24h", label="foo,bar=baz")
        self.assertEqual(
            client.call_args_list,
            [
                call(
                    "prune_containers",
                    filters={"until": ["24h"], "label": ["foo", "bar=baz"]},
                ),
                call(
                    "prune_images",
                    filters={"until": ["24h"], "label": ["foo", "bar=baz"]},
                ),
            ],
        )

        # System, no volumes, no filters, assuming prune_build in docker-py
        client = _run(system=True)
        self.assertEqual(
            client.call_args_list,
            [
                call("prune_containers", filters={}),
                call("prune_images", filters={}),
                call("prune_networks", filters={}),
                call("prune_build", filters={}),
            ],
        )

        # System, no volumes, no filters, assuming prune_build not in docker-py
        client = _run(
            system=True,
            side_effect=[None, None, None, SaltInvocationError(), None, None, None],
        )
        self.assertEqual(
            client.call_args_list,
            [
                call("prune_containers", filters={}),
                call("prune_images", filters={}),
                call("prune_networks", filters={}),
                call("prune_build", filters={}),
                call("_url", "/build/prune"),
                call("_post", None),
                call("_result", None, True),
            ],
        )

        # System, no volumes, with filters, assuming prune_build in docker-py
        client = _run(system=True, label="foo,bar=baz")
        self.assertEqual(
            client.call_args_list,
            [
                call("prune_containers", filters={"label": ["foo", "bar=baz"]}),
                call("prune_images", filters={"label": ["foo", "bar=baz"]}),
                call("prune_networks", filters={"label": ["foo", "bar=baz"]}),
                call("prune_build", filters={"label": ["foo", "bar=baz"]}),
            ],
        )

        # System, no volumes, with filters, assuming prune_build not in docker-py
        client = _run(
            system=True,
            label="foo,bar=baz",
            side_effect=[None, None, None, SaltInvocationError(), None, None, None],
        )
        self.assertEqual(
            client.call_args_list,
            [
                call("prune_containers", filters={"label": ["foo", "bar=baz"]}),
                call("prune_images", filters={"label": ["foo", "bar=baz"]}),
                call("prune_networks", filters={"label": ["foo", "bar=baz"]}),
                call("prune_build", filters={"label": ["foo", "bar=baz"]}),
                call("_url", "/build/prune"),
                call("_post", None),
                call("_result", None, True),
            ],
        )

        # System and volumes, no filters, assuming prune_build in docker-py
        client = _run(system=True, volumes=True)
        self.assertEqual(
            client.call_args_list,
            [
                call("prune_containers", filters={}),
                call("prune_images", filters={}),
                call("prune_networks", filters={}),
                call("prune_build", filters={}),
                call("prune_volumes", filters={}),
            ],
        )

        # System and volumes, no filters, assuming prune_build not in docker-py
        client = _run(
            system=True,
            volumes=True,
            side_effect=[
                None,
                None,
                None,
                SaltInvocationError(),
                None,
                None,
                None,
                None,
            ],
        )
        self.assertEqual(
            client.call_args_list,
            [
                call("prune_containers", filters={}),
                call("prune_images", filters={}),
                call("prune_networks", filters={}),
                call("prune_build", filters={}),
                call("_url", "/build/prune"),
                call("_post", None),
                call("_result", None, True),
                call("prune_volumes", filters={}),
            ],
        )

        # System and volumes with filters, assuming prune_build in docker-py
        client = _run(system=True, volumes=True, label="foo,bar=baz")
        self.assertEqual(
            client.call_args_list,
            [
                call("prune_containers", filters={"label": ["foo", "bar=baz"]}),
                call("prune_images", filters={"label": ["foo", "bar=baz"]}),
                call("prune_networks", filters={"label": ["foo", "bar=baz"]}),
                call("prune_build", filters={"label": ["foo", "bar=baz"]}),
                call("prune_volumes", filters={"label": ["foo", "bar=baz"]}),
            ],
        )

        # System and volumes, with filters, assuming prune_build not in docker-py
        client = _run(
            system=True,
            volumes=True,
            label="foo,bar=baz",
            side_effect=[
                None,
                None,
                None,
                SaltInvocationError(),
                None,
                None,
                None,
                None,
            ],
        )
        self.assertEqual(
            client.call_args_list,
            [
                call("prune_containers", filters={"label": ["foo", "bar=baz"]}),
                call("prune_images", filters={"label": ["foo", "bar=baz"]}),
                call("prune_networks", filters={"label": ["foo", "bar=baz"]}),
                call("prune_build", filters={"label": ["foo", "bar=baz"]}),
                call("_url", "/build/prune"),
                call("_post", None),
                call("_result", None, True),
                call("prune_volumes", filters={"label": ["foo", "bar=baz"]}),
            ],
        )

    def test_port(self):
        """
        Test docker.port function. Note that this test case does not test what
        happens when a specific container name is passed and that container
        does not exist. When that happens, the Docker API will just raise a 404
        error. Since we're using as side_effect to mock
        docker.inspect_container, it would be meaningless to code raising an
        exception into it and then test that we raised that exception.
        """
        ports = {
            "foo": {
                "5555/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32768"}],
                "6666/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32769"}],
            },
            "bar": {
                "4444/udp": [{"HostIp": "0.0.0.0", "HostPort": "32767"}],
                "5555/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32768"}],
                "6666/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32769"}],
            },
            "baz": {
                "5555/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32768"}],
                "6666/udp": [{"HostIp": "0.0.0.0", "HostPort": "32769"}],
            },
        }
        list_mock = MagicMock(return_value=["bar", "baz", "foo"])
        inspect_mock = MagicMock(
            side_effect=lambda x: {"NetworkSettings": {"Ports": ports.get(x)}}
        )
        with patch.object(docker_mod, "list_containers", list_mock), patch.object(
            docker_mod, "inspect_container", inspect_mock
        ):

            # Test with specific container name
            ret = docker_mod.port("foo")
            self.assertEqual(ret, ports["foo"])

            # Test with specific container name and filtering on port
            ret = docker_mod.port("foo", private_port="5555/tcp")
            self.assertEqual(ret, {"5555/tcp": ports["foo"]["5555/tcp"]})

            # Test using pattern expression
            ret = docker_mod.port("ba*")
            self.assertEqual(ret, {"bar": ports["bar"], "baz": ports["baz"]})
            ret = docker_mod.port("ba?")
            self.assertEqual(ret, {"bar": ports["bar"], "baz": ports["baz"]})
            ret = docker_mod.port("ba[rz]")
            self.assertEqual(ret, {"bar": ports["bar"], "baz": ports["baz"]})

            # Test using pattern expression and port filtering
            ret = docker_mod.port("ba*", private_port="6666/tcp")
            self.assertEqual(
                ret, {"bar": {"6666/tcp": ports["bar"]["6666/tcp"]}, "baz": {}}
            )
            ret = docker_mod.port("ba?", private_port="6666/tcp")
            self.assertEqual(
                ret, {"bar": {"6666/tcp": ports["bar"]["6666/tcp"]}, "baz": {}}
            )
            ret = docker_mod.port("ba[rz]", private_port="6666/tcp")
            self.assertEqual(
                ret, {"bar": {"6666/tcp": ports["bar"]["6666/tcp"]}, "baz": {}}
            )
            ret = docker_mod.port("*")
            self.assertEqual(ret, ports)
            ret = docker_mod.port("*", private_port="5555/tcp")
            self.assertEqual(
                ret,
                {
                    "foo": {"5555/tcp": ports["foo"]["5555/tcp"]},
                    "bar": {"5555/tcp": ports["bar"]["5555/tcp"]},
                    "baz": {"5555/tcp": ports["baz"]["5555/tcp"]},
                },
            )
            ret = docker_mod.port("*", private_port=6666)
            self.assertEqual(
                ret,
                {
                    "foo": {"6666/tcp": ports["foo"]["6666/tcp"]},
                    "bar": {"6666/tcp": ports["bar"]["6666/tcp"]},
                    "baz": {"6666/udp": ports["baz"]["6666/udp"]},
                },
            )
            ret = docker_mod.port("*", private_port="6666/tcp")
            self.assertEqual(
                ret,
                {
                    "foo": {"6666/tcp": ports["foo"]["6666/tcp"]},
                    "bar": {"6666/tcp": ports["bar"]["6666/tcp"]},
                    "baz": {},
                },
            )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.rabbitmq_cluster as rabbitmq_cluster

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class RabbitmqClusterTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate the rabbitmq_cluster state
    """

    def setup_loader_modules(self):
        return {rabbitmq_cluster: {}}

    def test_joined(self):
        """
        Test to ensure the current node joined
        to a cluster with node user@host
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

        mock = MagicMock(side_effect=[["rahulha@salt"], [""], [""]])
        with patch.dict(rabbitmq_cluster.__salt__, {"rabbitmq.cluster_status": mock}):
            ret.update({"comment": "Already in cluster"})
            self.assertDictEqual(
                rabbitmq_cluster.joined("salt", "salt", "rahulha"), ret
            )

            with patch.dict(rabbitmq_cluster.__opts__, {"test": True}):
                ret.update(
                    {
                        "result": None,
                        "comment": "Node is set to join " "cluster rahulha@salt",
                        "changes": {"new": "rahulha@salt", "old": ""},
                    }
                )
                self.assertDictEqual(
                    rabbitmq_cluster.joined("salt", "salt", "rahulha"), ret
                )

            with patch.dict(rabbitmq_cluster.__opts__, {"test": False}):
                mock = MagicMock(return_value={"Error": "ERR"})
                with patch.dict(
                    rabbitmq_cluster.__salt__, {"rabbitmq.join_cluster": mock}
                ):
                    ret.update({"result": False, "comment": "ERR", "changes": {}})
                    self.assertDictEqual(
                        rabbitmq_cluster.joined("salt", "salt", "rahulha"), ret
                    )



# -*- coding: utf-8 -*-
"""
Pillar include tests
"""
from __future__ import absolute_import, unicode_literals

from tests.support.case import ModuleCase


class PillarIncludeTest(ModuleCase):
    def test_pillar_include(self):
        """
        Test pillar include
        """
        ret = self.minion_run("pillar.items")
        assert "a" in ret["element"]
        assert ret["element"]["a"] == {"a": ["Entry A"]}
        assert "b" in ret["element"]
        assert ret["element"]["b"] == {"b": ["Entry B"]}

    def test_pillar_glob_include(self):
        """
        Test pillar include via glob pattern
        """
        ret = self.minion_run("pillar.items")
        assert "glob-a" in ret
        assert ret["glob-a"] == ["Entry A"]
        assert "glob-b" in ret
        assert ret["glob-b"] == ["Entry B"]



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.drac as drac

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class DracTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.drac
    """

    def setup_loader_modules(self):
        return {drac: {}}

    # 'present' function tests: 1

    def test_present(self):
        """
        Test to ensure the user exists on the Dell DRAC
        """
        name = "damian"
        password = "secret"
        permission = "login,test_alerts,clear_logs"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        mock = MagicMock(return_value=[name])
        with patch.dict(drac.__salt__, {"drac.list_users": mock}):
            with patch.dict(drac.__opts__, {"test": True}):
                comt = "`{0}` already exists".format(name)
                ret.update({"comment": comt})
                self.assertDictEqual(drac.present(name, password, permission), ret)

            with patch.dict(drac.__opts__, {"test": False}):
                comt = "`{0}` already exists".format(name)
                ret.update({"comment": comt})
                self.assertDictEqual(drac.present(name, password, permission), ret)

    # 'absent' function tests: 1

    def test_absent(self):
        """
        Test to ensure a user does not exist on the Dell DRAC
        """
        name = "damian"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        mock = MagicMock(return_value=[])
        with patch.dict(drac.__salt__, {"drac.list_users": mock}):
            with patch.dict(drac.__opts__, {"test": True}):
                comt = "`{0}` does not exist".format(name)
                ret.update({"comment": comt})
                self.assertDictEqual(drac.absent(name), ret)

            with patch.dict(drac.__opts__, {"test": False}):
                comt = "`{0}` does not exist".format(name)
                ret.update({"comment": comt})
                self.assertDictEqual(drac.absent(name), ret)

    # 'network' function tests: 1

    def test_network(self):
        """
        Test to ensure the DRAC network settings are consistent
        """
        ip_ = "10.225.108.29"
        netmask = "255.255.255.224"
        gateway = "10.225.108.1"

        ret = {"name": ip_, "result": None, "comment": "", "changes": {}}

        net_info = {
            "IPv4 settings": {
                "IP Address": ip_,
                "Subnet Mask": netmask,
                "Gateway": gateway,
            }
        }

        mock_info = MagicMock(return_value=net_info)
        mock_bool = MagicMock(side_effect=[True, False])
        with patch.dict(
            drac.__salt__,
            {"drac.network_info": mock_info, "drac.set_network": mock_bool},
        ):
            with patch.dict(drac.__opts__, {"test": True}):
                self.assertDictEqual(drac.network(ip_, netmask, gateway), ret)

            with patch.dict(drac.__opts__, {"test": False}):
                comt = "Network is in the desired state"
                ret.update({"comment": comt, "result": True})
                self.assertDictEqual(drac.network(ip_, netmask, gateway), ret)

                comt = "unable to configure network"
                ret.update({"comment": comt, "result": False})
                self.assertDictEqual(drac.network(ip_, netmask, gateway), ret)



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
import salt.modules.glance as glance
import salt.modules.salt_version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch

# Import Salt Testing libs
from tests.support.unit import TestCase


class GlanceTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {
            glance: {
                "__salt__": {
                    "salt_version.less_than": salt.modules.salt_version.less_than
                }
            },
        }

    def test_image_list(self):
        """
        test salt.modles.glance
        """
        name = "test"
        image = MagicMock()
        image.name = name
        attrs = {
            "images.list.return_value": [image],
        }
        mock_auth = MagicMock(**attrs)
        patch_auth = patch("salt.modules.glance._auth", return_value=mock_auth)

        with patch_auth:
            ret = glance.image_list(id="test_id", name=name)
            assert ret[0]["name"] == name
